//
// Convergence seems to be reasonable but many iterations are usually
// required if the initial guess is bad.
//
// 
// by Carlos E. Christoffersen

#include <carrot/Circuit.h>
#include <carrot/OP.h>
#include <carrot/Options.h>
#include "DC.h"
#include <carrot/UMFPackDI.h>
#include "global.h"

// For now put the include here to test the routines
#include <carrot/DoubleHolder.h>

#include <stdlib.h>
#include <cmath>
#include <string>
#include <iostream>

// Debug
// #define DEBUG 1

#ifdef DEBUG
#include <cstdio>
#endif

static ItemInfo DC_info = {
  "dc",
  "DC Analysis",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

enum {SSTART, SSTOP, SSTEP, MAX_DPARM_INDEX}; 

static ParamHandler<double>::ParmInfo DC_dparams[] = {
  {"start", "Start value", true, 0.},
  {"stop", "Stop value", true, 0.},
  {"step", "Step value", true, 0.}
};

// ***********************************************************************
// Integer parameters
// ***********************************************************************
enum {NSTEPS, MAX_IPARM_INDEX};

static ParamHandler<int>::ParmInfo DC_iparams[] = {
  {"nsteps", "Number of steps to skip in progress output", false, 100}
};

// ***********************************************************************
// String parameters
// ***********************************************************************
enum {DEVNAME, PARNAME, MAX_SPARM_INDEX};

static ParamHandler<std::string>::ParmInfo DC_sparams[] = {
  {"dev", "Device name", true, "xx"},
  {"par", "Parameter to sweep", true, "xx"}
};

extern "C" {

  //Export the name of the module we are using
  const char * AnalysisName(){
    return DC_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model, &DC_info, 
		       DC_dparams, MAX_DPARM_INDEX, 
		       DC_iparams, MAX_IPARM_INDEX, 
		       DC_sparams, MAX_SPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }
  Analysis * newObject(ParmSet * p){
    return new DC(p);
}
  void delObject(Analysis * e){
    delete e;
    e = 0;
  }
};

void DC::run(Circuit* cir, OutputManager* om)
{
  // Check parameters
  getParmSet()->checkParms();

  // Set variables in OutputManager
  om->setVarUnit("-"); // For now no units if no sweep
  om->setAnalysis("dc");

  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  std::string *sparm = getParmSet()->getSHandler()->getParmVec();

  printf("****************************************************************\n");
  printf("*                          DC Analysis                         *\n");
  printf("****************************************************************\n");

  // Set row/column numbers in terminals and get number of DC-used
  // terminals. Linear and nonlinear charge sources are excluded to
  // avoid a singular matrix when a node is not resistively connected
  // to ground. In this pathological case the node numbering is not
  // the same as when all elements are included.
  int nterms = cir->setMatrixRCs();
  // Get rows of desired output variables
  std::vector<int> iv;
  om->getRowNumbers(cir, iv);
  // ***********************  Create matrix and vectors

  // Ground node not included in nrcs
  int nrcs = nterms - 1; 
  // NAM matrix. No need to clean matrix because it is fresh.
  UMFPackDI nam(nrcs, 24*nrcs); 

  // Right hand side vector (source vector)
  double *svec_ground = new double[nrcs+1]; 
  double *svec = svec_ground + 1;
  // unknowns vector (with ground)
  double *uvec_ground = new double[nrcs+1]; 
  // Ground node potential is zero
  uvec_ground[0] = 0.; 
  // This would be uvec_(j+1)
  double *uvec = uvec_ground + 1;

  // Set initial guess to zero
  for (int i(0); i<nrcs; i++)
    uvec[i] = 0.;
  
  // Search device in circuit and check that parameter exists and is
  // double type
  MacroElement* me = cir->getMElement(sparm[DEVNAME]);
  if (!me) {
    // Device does not exist
    printf("Fatal: Could not find device in netlist: \"%s\".\n", 
	   sparm[DEVNAME].c_str());
    exit(EXIT_FAILURE);
  }
  // Now search parameter index (if it exists)
  int max_idx = me->getParmSet()->getDHandler()->getNParms();
  int par_idx = me->getParmSet()->getDHandler()->getIndex(sparm[PARNAME]);
  if (par_idx >= max_idx) {
    // Device does not exist
    printf("Fatal: Could not find double parameter named \"%s\" in device \"%s\".\n", 
	   sparm[PARNAME].c_str(), sparm[DEVNAME].c_str());
    exit(EXIT_FAILURE);
  }
  // If we reached this point, the device and parameters are OK.
  // Save current parameter value
  double saved_value = me->getParmSet()->getDHandler()->getParmVec()[par_idx];
  OP opoint(cir);
  int iter, rejections;

  printf("----------------------------------------------------------------------\n");
  printf(" Step \t %s \t Iterations \t Rejections \t Residual \n", 
	 sparm[PARNAME].c_str());
  printf("----------------------------------------------------------------------\n");
  // For now do a linear sweep only
  int counter = 1;
  double signstep = dparm[SSTEP] / fabs(dparm[SSTEP]);
  for (double value = dparm[SSTART]; 
       value * signstep < dparm[SSTOP] * signstep; 
       value += dparm[SSTEP], counter++) {
    // Set parameter value
    me->getParmSet()->getDHandler()->setParm(par_idx, value);
    // Re-initialize current sources with new parameter
    me->initAllCS();
    // Initial guess is set to zero at first, and later is set to
    // previous solution.
    double residual = opoint.op(nrcs, nam, uvec, svec, iter, rejections);
    // Clear nam
    nam.reset();
    // Skip some steps to avoid fast scrolling
    if (!(counter % iparm[NSTEPS]) || counter==1) {
      printf(" %d \t %g \t      %d \t       %d \t  %g\n", 
	     counter, value, iter, rejections, residual);
    }
    if (residual > options.getDHandler()->getParmVec()[OPT_TOL]) {
      printf("WARNING: Step: %d, %s: %g, Convergence not achieved.\n", 
	     counter, sparm[PARNAME].c_str(), value);
      printf("WARNING: Residual = %g (Results possibly wrong.)\n\n",
	     residual);
    }
    // ****************   Save results here 
    // om takes care of cleaning this ...
    DoubleHolder* dh = new DoubleHolder(iv.size()); 
    // Set sweep variable value
    dh->setVal(value);
    int iend = iv.size();
    for(int i(0); i<iend; i++) 
      dh->saveValue(i, uvec[iv[i]]);
    om->addSample(dh);
  }
  printf("----------------------------------------------------------------------\n");
  // Restore parameter value
  me->getParmSet()->getDHandler()->setParm(par_idx, saved_value);
  me->initAllCS();

  // Clean up everything (in reverse order they were created)
  delete [] uvec_ground;
  delete [] svec_ground;

  // nam gets destroyed as it goes out of scope
  printf("********************** End Of DC Analysis **********************\n");
}


