#include <carrot/NewtonSolve.h>
#include "TimeMarching.h"
#include "BEuler.h"
#include "Trapezoidal.h"

#include <cmath>

template <typename IntegMethod>
TimeMarching<IntegMethod>::TimeMarching(Circuit* cir, const int& nrcs, 
					IntegMethod* imo, UMFPackDI& nam)
  : cir(cir), nrcs(nrcs), imo(imo)
{
  // Get the element vectors we need first
  linvccs = cir->getLinVCCSvec();
  genvcqs = cir->getGenVCQSvec();
  linvcqs = cir->getLinVCQSvec();
  genvccs = cir->getGenVCCSvec();
  indvccs = cir->getIndepCSvec();

  // Number of nonlinear charges in circuit
  const int& nCharges = genvcqs.size();

  // Independent source contributions
  indep_svec = new double[nrcs];

  // Get the maximum number of controlling voltages we expect
  const int &max_cv = cir->getMaxCV();

  // **********************  Work space vectors
  dij = new double[max_cv];
  dqj = new double[max_cv];
  
  vj = new double[max_cv];
  q = new double[nCharges + 1];

  // Consider linear VCCSs: These never change (no time-varying linear
  // elements for now). We create this part of the matrix in triplet
  // form just once. Then we add/erase only the nonlinear part at each
  // iteration.
  for(LinVCCSVector::iterator i = linvccs.begin(); i < linvccs.end(); i++ )
    {
      double g = (*i) -> getGain();
      nam.addElement((*i)-> getRow1(), (*i)->getCol1(), g );
      nam.addElement((*i)-> getRow1(), (*i)->getCol2(), -g );
      nam.addElement((*i)-> getRow2(), (*i)->getCol1(), -g );
      nam.addElement((*i)-> getRow2(), (*i)->getCol2(), g );
    }

  // Mark the endof contribution of linear elements in matrix
  nam.setGMark();
}

template <typename IntegMethod>
TimeMarching<IntegMethod>::~TimeMarching()
{
  delete [] q;
  delete [] vj;
  delete [] dqj;
  delete [] dij;
  delete [] indep_svec;
}

template <typename IntegMethod>
bool TimeMarching<IntegMethod>::step(UMFPackDI& nam, double* uvec, 
				     double* q, double* svec, 
				     const double& t, const double& tstep, 
				     const bool& newtstep,
				     double& residual, int& iterations)
{
  // Erase source vector.
  for (int i(0); i<nrcs; i++){
    svec[i] = 0.;
  }
  // Consider independent CS.
  for( IndepCSVector::iterator i = indvccs.begin(); i < indvccs.end(); i++ )
    {
      double ivalue;
      (*i) -> eval(ivalue, t+tstep );
      svec[(*i) -> getRow1()] -= ivalue;
      svec[(*i) -> getRow2()] += ivalue;
    }

  // Assume the time step may change
  if (newtstep) {
    imo->seth(tstep);
    //****************************************************************
    //*************************CAPACITOR******************************
    //****************************************************************
    nam.clearCapNumeric();
    const double& cFactor(imo->getCfactor()); 
    for(LinVCQSVector::iterator i = linvcqs.begin(); 
	i < linvcqs.end(); i++ )
      {
	double geq = (*i)->getGain() * cFactor;
	nam.addElement((*i)-> getRow1(), (*i)->getCol1(), geq );
	nam.addElement((*i)-> getRow1(), (*i)->getCol2(), -geq );
	nam.addElement((*i)-> getRow2(), (*i)->getCol1(), -geq );
	nam.addElement((*i)-> getRow2(), (*i)->getCol2(), geq );
      }
    nam.setLinMark();
  }
  for(LinVCQSVector::iterator i = linvcqs.begin(); i < linvcqs.end(); i++ )
    {
      double ieq = imo->getCieq((*i)->getCol1(), (*i)->getCol2(), 
			       (*i)->getGain());
      svec[(*i)->getRow1()] += ieq;
      svec[(*i)->getRow2()] -= ieq;
    }

      
  // Save invariant part of source vector.
  for( int i(0); i <nrcs; i++ ) {
    indep_svec[i] = svec[i];
  }
  // *********************************************************************
  // Solve nonlinear problem
  // *********************************************************************
  int rejections;
  NewtonSolve< TimeMarching<IntegMethod> > nw(this, &nam, nrcs);
  residual = nw.solve(uvec, indep_svec, iterations, rejections);

  // Save the charge values corresponding to the final nodal voltages
  const int& nCharges = genvcqs.size();
  for (int i(0); i < nCharges; i++) {
    // Calculate vj.
    for (int k(0); k < genvcqs[i]->getNCVolt(); k++) {
      // uvec_j[-1] = 0 (takes care of ground node)
      vj[k] = uvec[genvcqs[i]->getColPlus()[k]]
	- uvec[genvcqs[i]->getColMinus()[k]];
    }
    // Evaluate nonlinear element
    genvcqs[i]->eval(q[i], vj);
  }

  // ------------------------------------------------------
  // Fix this later
  if (residual > 1e-6)
    return false;
  else
    return true;
}

template <typename IntegMethod>
int TimeMarching<IntegMethod>::fillNAMandRHS(UMFPackDI& nam, double* svec, 
					      double* uvec_j, 
					      const double& beta, 
					      const double& gmin)
{
  const int& nCharges = genvcqs.size();
  int retcode = 1; // 1 means linear problem
  if (genvccs.size() || nCharges)
    retcode = 0; // 0 means normal nonlinear problem

  // Consider now generic VCCSs
  for (unsigned i(0); i < genvccs.size(); i++) {

    // Calculate vj.
    for (int k(0); k < genvccs[i]->getNCVolt(); k++) {
      // uvec_j[-1] = 0 (takes care of ground node)
      vj[k] = uvec_j[genvccs[i]->getColPlus()[k]]
	- uvec_j[genvccs[i]->getColMinus()[k]];
    }


    // Evaluate nonlinear element
    double ij;
    genvccs[i]->eval_and_deriv(ij, dij, vj);

    // Fill source vector. Current flows from row 1 to 2.
    svec[genvccs[i]->getRow1()] -= ij;
    svec[genvccs[i]->getRow2()] += ij;

    for (int k(0); k < genvccs[i]->getNCVolt(); k++) {

      // svec[-1] is allocated but not used anywhere
      svec[genvccs[i]->getRow1()] += dij[k] * vj[k] * beta;
      svec[genvccs[i]->getRow2()] -= dij[k] * vj[k] * beta;

      // Fill NAM
      double geq = dij[k] * beta + gmin;
      nam.addElement(genvccs[i]->getRow1(), genvccs[i]->getColPlus()[k],
		     geq);
      nam.addElement(genvccs[i]->getRow1(), genvccs[i]->getColMinus()[k],
		     -geq);
      nam.addElement(genvccs[i]->getRow2(), genvccs[i]->getColPlus()[k],
		     -geq);
      nam.addElement(genvccs[i]->getRow2(), genvccs[i]->getColMinus()[k],
		     geq);
    }
  }

  // Consider now generic VCQSs
  for (int i(0); i < nCharges; i++) {

    // Calculate vj.
    for (int k(0); k < genvcqs[i]->getNCVolt(); k++) {
      // uvec_j[-1] = 0 (takes care of ground node)
      vj[k] = uvec_j[genvcqs[i]->getColPlus()[k]]
	- uvec_j[genvcqs[i]->getColMinus()[k]];
    }


    // Evaluate nonlinear element
    double qj;
    genvcqs[i]->eval_and_deriv(qj, dqj, vj);
    double ieq;
    imo->getIeqQ(qj, i, ieq);
    // svec[-1] is allocated but not used anywhere
    svec[genvcqs[i]->getRow1()] += ieq;
    svec[genvcqs[i]->getRow2()] -= ieq;

    for (int k(0); k < genvcqs[i]->getNCVolt(); k++) {

      double geq;
      imo->getIeqGeq(dqj[k] * beta, vj[k], ieq, geq);

      // svec[-1] is allocated but not used anywhere
      svec[genvcqs[i]->getRow1()] += ieq;
      svec[genvcqs[i]->getRow2()] -= ieq;

      // Fill NAM
      geq += gmin;
      nam.addElement(genvcqs[i]->getRow1(), genvcqs[i]->getColPlus()[k],
		     geq);
      nam.addElement(genvcqs[i]->getRow1(), genvcqs[i]->getColMinus()[k],
		     -geq);
      nam.addElement(genvcqs[i]->getRow2(), genvcqs[i]->getColPlus()[k],
		     -geq);
      nam.addElement(genvcqs[i]->getRow2(), genvcqs[i]->getColMinus()[k],
		     geq);

    }
  }
  return retcode;
}


// Explicitly instantiate templates here
template class TimeMarching<BEuler>;
template class TimeMarching<Trapezoidal>;
