/* 
   carrot - Circuit Simulator Program

   Copyright (C) 2004 Carlos Christoffersen

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#include <map>
#include <string>
#include <iostream>
#include <fstream>
#include <carrot/Circuit.h>
#include <carrot/Analysis.h>
#include "analysis/AnalysisLoader.h"

#include <cstdio>
#include <sys/types.h>
#include <getopt.h>
#include "system.h"

#define EXIT_FAILURE 1

extern "C" {
  char *xstrdup (char *p);
}

static void usage (int status);

/* The name the program was run with, stripped of any leading path. */
char *program_name;

/* Option flags and variables */
static struct option const long_options[] =
{
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'V'},
  {NULL, 0, NULL, 0}
};

static int decode_switches (int argc, char **argv);

// Defined in parser.cc
void parse(Circuit &cir, 
	   std::map<std::string, ParmSet *> &analysisParameters,
	   AnalysisMap &analysisLibraries, std::istream *file_p);

// Flag to indicate if we want to print shared library loading for devices
bool printlib(false);

int
main (int argc, char **argv)
{
  int i;
  program_name = argv[0];

  i = decode_switches (argc, argv);

  printf("\n----------------------------------------------------------------------------\n");
  printf("+++                    Carrot Circuit Simulator                          +++\n");
  printf("----------------------------------------------------------------------------\n\n");

  // Open input file name if specified. 
  std::istream *file_p;
  std::ifstream input_F;
  if(argc > i) {
    // argv[i] must be the file name
    input_F.open(argv[i]);
    if (!input_F.good()) {
      fprintf(stderr,"Error: couldn't open input file: %s\n", argv[i]);
      exit(1);
    }
    file_p = &input_F;
    printf("Input file: %s\n\n", argv[i]);  
  } 
  else {
    printf("Using standard input.\n\n");
    file_p = &(std::cin); 
  }

  if (!printlib) {
    printf("Loading shared libraries ... ");
    fflush(stdout);
  }
  // Create circuit (device shared libraries open inside Circuit class)
  Circuit cir(printlib);
  // Create analysis map
  std::map<std::string, ParmSet *> analysisParameters;
  AnalysisMap analysisLibraries;
  // Open analysis shared libraries here
  analysisLibraries = openAnalysis(ANALYSIS_DIR_PATH, printlib);
  if (!printlib)
    printf("done.\n\n");

  // Parse the circuit
  printf("Parsing netlist ... ");
  fflush(stdout);
  parse(cir, analysisParameters, analysisLibraries, file_p);
  printf("done.\n\n");

  // Initialise circuit (add any checks later)
  cir.init();

  printf("----------------------------------------------------------------------------\n\n");

  // Print circuit if requested
  cir.print();

  //Start simulation here 

  // Run the requested analyses
  std::map<std::string, ParmSet *>::iterator runAnalysis;
  runAnalysis = analysisParameters.begin();
  Analysis * analysisObject = 0;
  while(runAnalysis != analysisParameters.end() ){
    // Create a new OutputManager per analysis run
    OutputManager* om = new OutputManager();
    // Create analysis object, run and free object
    analysisObject = 
      analysisLibraries[runAnalysis->first].newObject(runAnalysis->second);
    analysisObject->run(&cir, om);
    // Later add output processing here
    analysisLibraries[runAnalysis->first].delObject(analysisObject);
    analysisObject = 0;
    // Write analysis output
    om->write();
    delete om;
    ++runAnalysis;
  }
  closeAnalysis(analysisLibraries);

  exit (0);
}

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */

static int
decode_switches (int argc, char **argv)
{
  int c;


  while ((c = getopt_long (argc, argv, 
			   "h"	/* help */
			   "V"	/* version */
			   "s",  // Print device library loading
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'V':
	  printf ("carrot %s\n", VERSION);
	  exit (0);

	case 's':
	  printlib = true;
	  continue;
	  
	case 'h':
	  usage (0);

	default:
	  usage (EXIT_FAILURE);
	}
    }

  return optind;
}


static void
usage (int status)
{
  printf (_("\n%s - \
Circuit Simulator Program\n\n"), program_name);
  printf ("Usage: %s [OPTION]... [INPUT_FILE]...\n", program_name);
  printf ("\nIf [INPUT_FILE] is not specified, read from standard input.\n");

  printf ("\nOptions:\n\n");
  printf ("-h, --help                 display this help and exit\n");
  printf ("-V, --version              output version information and exit\n\n");
  printf ("-s                         verbose shared library printout\n");
  exit (status);
}
