#include <carrot/Circuit.h>
#include <carrot/ObjectLoader.h>
#include <carrot/Options.h>

#include <iostream>
#include <cstdio>


Circuit::Circuit(bool libflag) : psetMap(), melemMap(), termMap(),
				 lCVec(), flCVec(), iCVec(), gCVec(), tdCVec(),
				 lQVec(), gQVec(), mVec(), tVec(), 
				 init_flag(false), max_dparms(0), 
				 max_cv(0), max_cs(0), uptodate(false)
{
  // Get macroelement definitions from shared libraries
  elementLibraryMap = openModules(DEVS_DIR_PATH, libflag);
  // The circuit is empty. Just the ground terminal is present.
  Terminal *gnd = new Terminal("gnd");
  termMap[gnd->getName()] = gnd;
  tVec.add(gnd);
}

Circuit::~Circuit()
{
  // We can not delete all at once if macroelements clean after
  // themselves and know nothing about the circuit class. We have no
  // choice but to remove MacroElements one by one. Then we can delete
  // any remaining CS.
  MacroElVector& mEvec = mVec.getVec();
  MacroElVector::iterator iend = mEvec.end();
  for (MacroElVector::iterator i = mEvec.begin(); i < iend; i++)
    if (*i)
      remove(*i);

  // Clean any CS that is still there
  lCVec.eraseElements();
  flCVec.eraseElements();
  iCVec.eraseElements();
  gCVec.eraseElements();
  tdCVec.eraseElements();
  lQVec.eraseElements();
  gQVec.eraseElements();
  // Delete any remaining terminals
  tVec.eraseElements();

  // Delete parameter sets
  ParmSetMap::iterator psetend = psetMap.end();
  for (ParmSetMap::iterator i(psetMap.begin()); i != psetend; i++) {
    //    delete (*i).second;
    // Delete using library function to be clean
    elementLibraryMap[(*i).second->getInfo()->type].delParmSet((*i).second);
  }
  closeModules(elementLibraryMap);
}

void Circuit::addParmSet(ParmSet* pset)
{
  // Make sure the parameter set is not already in the circuit.
  assert(psetMap.find(pset->getName()) == psetMap.end());
  psetMap[pset->getName()] = pset;
  int nparms = pset->getDHandler()->getNParms();
  if (nparms > max_dparms)
    max_dparms = nparms;
}

ParmSet* Circuit::getParmSet(const std::string& name)
{
  ParmSetMap::iterator i = psetMap.find(name);
  if (i != psetMap.end())
    return (*i).second;
  else
    return NULL;
}

void Circuit::addElement(LinearVCCS* vccs)
{
  lCVec.add(vccs);
  uptodate = false;
}

void Circuit::remove(LinearVCCS* vccs)
{
  lCVec.remove(vccs);
  disconnect(vccs);
  uptodate = false;
}

void Circuit::addElement(LinearfVCCS* vccs)
{
  flCVec.add(vccs);
  uptodate = false;
}

void Circuit::remove(LinearfVCCS* vccs)
{
  flCVec.remove(vccs);
  disconnect(vccs);
  uptodate = false;
}

void Circuit::addElement(LinearVCQS* vccs)
{
  lQVec.add(vccs);
  uptodate = false;
}

void Circuit::remove(LinearVCQS* vcqs)
{
  lQVec.remove(vcqs);
  disconnect(vcqs);
  uptodate = false;
}

void Circuit::addElement(IndependentCS* ics)
{
  iCVec.add(ics);
  uptodate = false;
}

void Circuit::remove(IndependentCS* ics)
{
  iCVec.remove(ics);
  disconnect(ics);
  uptodate = false;
}

void Circuit::addElement(GenericVCCS* vccs)
{
  gCVec.add(vccs);
  uptodate = false;
}

void Circuit::remove(GenericVCCS* vccs)
{
  gCVec.remove(vccs);
  disconnect(vccs);
  uptodate = false;
}

void Circuit::addElement(TimeDelayedCS* vccs)
{
  tdCVec.add(vccs);
  uptodate = false;
}

void Circuit::remove(TimeDelayedCS* vccs)
{
  tdCVec.remove(vccs);
  disconnect(vccs);
  uptodate = false;
}

void Circuit::addElement(GenericVCQS* vcqs)
{
  gQVec.add(vcqs);
  uptodate = false;
}

void Circuit::remove(GenericVCQS* vcqs)
{
  gQVec.remove(vcqs);
  disconnect(vcqs);
  uptodate = false;
}

void Circuit::addElement(MacroElement* me)
{
  // Make sure the element name is is not already in the circuit.
  assert(melemMap.find(me->getName()) == melemMap.end());
  // Add to map 
  melemMap[me->getName()] = me;
  mVec.add(me);
  uptodate = false;
}

template <typename VEC>
void Circuit::eraseCSVector(VEC& vec)
{
  for (int jend(vec.size()), j(0); j < jend; j++) {
    // Disconnect from terminals and remove from circuit (do not
    // delete object, that is performed in the MacroElement
    // Destructor).
    remove(vec[j]);
  }
}

void Circuit::remove(MacroElement* me)
{
  // Remove from map
  melemMap.erase(me->getName());
  // Remove from vector
  mVec.remove(me);
  // Delete internal current sources (and terminals). 
  eraseCSVector(me->getLinearVCCSvec());
  eraseCSVector(me->getLinearfVCCSvec());
  eraseCSVector(me->getLinearVCQSvec());
  eraseCSVector(me->getGenVCCSvec());
  eraseCSVector(me->getTDCSvec());
  eraseCSVector(me->getGenVCQSvec());
  eraseCSVector(me->getIndepCSvec());
  // Remove links from external terminals
  disconnect(me);
  // delete MacroElement itself
  //  delete me;
  // Delete using library function to be clean
  elementLibraryMap[me->getParmSet()->getInfo()->type].delObject(me);
  // Internal terminals are left floating, remove with:
  // removeFloatingTerms() (not practical to use here because of too
  // much wasted time if many elements are removed).
  uptodate = false;
}

void Circuit::disconnect(ElementNode* en)
{
  int nterms = en->getNTerms();
  for (int i(0); i < nterms; i++) {
    Terminal *term = en->getTerminal(i);
    if (term)
      term->unlink(en);
  }
}

void Circuit::removeFloatingTerms()
{
  TerminalVector& termvec = tVec.getVec();
  TerminalVector::iterator iend = termvec.end();
  for (TerminalVector::iterator i = termvec.begin(); i < iend; i++)
    if ((*i)->getNConnections() == 0)
      remove(*i);
}

MacroElement* Circuit::getMElement(const std::string& name)
{
  MElemMap::iterator i = melemMap.find(name);
  if (i != melemMap.end())
    return (*i).second;
  else
    return NULL;
}

void Circuit::addTerminal(Terminal* term)
{
  // Make sure the terminal name is not already in the circuit.
  assert(termMap.find(term->getName()) == termMap.end());
  termMap[term->getName()] = term;
  tVec.add(term);
  // Add to map too
  uptodate = false;
}

void Circuit::remove(Terminal* term)
{
  // Remove from map
  termMap.erase(term->getName());
  // Remove from vector
  tVec.remove(term);
  uptodate = false;
}

Terminal* Circuit::getTerminal(const std::string& name)
{
  TermMap::iterator i = termMap.find(name);
  if (i != termMap.end())
    return (*i).second;
  else
    return NULL;
}


template <typename VEC>
void Circuit::processCSVector(VEC& vec, 
			      TerminalVector& metv, const int& totTerms)
{
  for (int jend(vec.size()), j(0); j < jend; j++) {
    // Insert CS
    addElement(vec[j]);
    int csterms = vec[j]->getNTerms();
    const int* nodevec = vec[j]->getTermNumVec();
    for (int k(0); k < csterms; k++) {
      //      std::cout << nodevec[k] << std::endl;
      assert(nodevec[k] < totTerms);
      // Connect terminal
      if (nodevec[k] > -1)
	vec[j]->link(k, metv[nodevec[k]]);
      else
	// we are dealing with the ground node
	vec[j]->link(k, tVec.getElem(0));
    }
  }
}

void Circuit::init()
{
  MacroElVector& mEvec = mVec.getVec();
  // Go through all MacroElements and call init() 
  MacroElVector::iterator iend = mEvec.end();
  for (MacroElVector::iterator i = mEvec.begin(); i < iend; i++) 
    initMacroElement(*i);
  // Consistency check:
  // Here we should check that all elements referenced by terminals
  // and all terminals referenced by elements are properly included in
  // the circuit.
  // consistencyCheck();
  init_flag = true;
}

void Circuit::initMacroElement(MacroElement* me)
{
  // Make sure that me is in the Circuit
  assert(me == mVec.getElem(me->getIndex()));
  // Expand MacroElement into CS elements
  me->init();
  //    std::cout << me->getName() << std::endl;
  int nterms = me->getNTerms();
  int totTerms = nterms + me->getNInternalNodes();
  TerminalVector metv(totTerms);
  //    std::cout << totTerms << std::endl;
  for (int j(0); j < nterms; j++)
    metv[j] = me->getTerminal(j);
  // Create internal terminals
  for (int j(nterms); j < totTerms; j++) {
    char buffer[255];
    sprintf(buffer, "%s:%s_%d", 
	    me->getParmSet()->getInfo()->type.c_str(), 
	    me->getName().c_str(), j);
    // Create Terminal (not possible to duplicate names with current
    // naming scheme). Set the internal flag to true.
    metv[j] = new Terminal(buffer, true);
    addTerminal(metv[j]);
  }
  processCSVector(me->getLinearVCCSvec(), metv, totTerms);
  processCSVector(me->getLinearfVCCSvec(), metv, totTerms);
  processCSVector(me->getLinearVCQSvec(), metv, totTerms);
  processCSVector(me->getGenVCCSvec(), metv, totTerms);
  processCSVector(me->getTDCSvec(), metv, totTerms);
  processCSVector(me->getGenVCQSvec(), metv, totTerms);
  processCSVector(me->getIndepCSvec(), metv, totTerms);
  // Initialise CS elements. 
  me->initAllCS();
}



template <typename T>
void Circuit::setTermTable(int *termTable, int& nterms, T& vec)
{
  int imax(vec.size());
  for (int i(0); i < imax; i++) {
    int nterms_this = vec[i]->getNTerms();
    for (int j(0); j < nterms_this; j++) {
      int cir_tn = vec[i]->getTerminal(j)->getIndex();
      // Add new row/column number if not in termTable (do not count ground)
      if (cir_tn && !termTable[cir_tn])
	termTable[cir_tn] = nterms++; 
    }
  }
}

template <typename T>
void Circuit::setRCInfo(T& vec)
{
  int imax(vec.size());
  for (int i(0); i < imax; i++) 
    vec[i]->setRCinfo();
}

void Circuit::setRCInfo(GenVCCSVector& vec) // Specialisation
{
  int imax(vec.size());
  for (int i(0); i < imax; i++) {
    vec[i]->setRCinfo();
    int ncvolt = vec[i]->getNCVolt();
    if (ncvolt > max_cv)
      max_cv = ncvolt;
  }
}

void Circuit::setRCInfo(TDCSVector& vec) // Specialisation
{
  int imax(vec.size());
  for (int i(0); i < imax; i++) {
    vec[i]->setRCinfo();
    int ncvolt = vec[i]->getNCVolt();
    if (ncvolt > max_cv)
      max_cv = ncvolt;
  }
}

int Circuit::setMatrixRCs()
{
  // Make sure there are no floating terminals left over.
  removeFloatingTerms();
  MacroElVector& mEvec = mVec.getVec();
  MacroElVector::iterator iend = mEvec.end();
  max_cs = 0;
  for (MacroElVector::iterator i = mEvec.begin(); i < iend; i++) {
    // Get the maximum number of current sources of a MacroElement
    int ncs = (*i)->getCSNumber();
    if (ncs > max_cs)
      max_cs = ncs;
  }
  
  int cir_nterms = getNumberOfTerms();
  int *termTable = new int[cir_nterms];

  // At first, 0 means the ground node. Then we decrement all elements
  // in termTable[] so the ground node becomes (-1).
  for (int i(0); i<cir_nterms; i++)
    termTable[i] = 0; 
  // Number of terminals, ground node included.
  int nterms(1); 

  // Now mark in termTable what terminals are actually used and the row
  // and assign row and column numbers (Check for each type of CS).
  LinVCCSVector& linVCCSvec = lCVec.getVec();
  LinfVCCSVector& linfVCCSvec = flCVec.getVec();
  LinVCQSVector& linVCQSvec = lQVec.getVec();
  GenVCCSVector& genVCCSvec = gCVec.getVec();
  TDCSVector& tdCSvec = tdCVec.getVec();
  GenVCQSVector& genVCQSvec = gQVec.getVec();
  IndepCSVector& indepCSvec = iCVec.getVec();

  setTermTable(termTable, nterms, linVCCSvec);
  setTermTable(termTable, nterms, linfVCCSvec);
  setTermTable(termTable, nterms, linVCQSvec);
  setTermTable(termTable, nterms, genVCCSvec);
  setTermTable(termTable, nterms, tdCSvec);
  setTermTable(termTable, nterms, genVCQSvec);
  setTermTable(termTable, nterms, indepCSvec);

  // Set Terminal RCs. Decrement all RCs numbers by one to use
  // zero-indexing in matrix.  -1: ground node
  TerminalVector& termvec = tVec.getVec();
  for (int i(0); i<cir_nterms; i++)
    termvec[i]->setRC(termTable[i] - 1);

  // Now setup quick-access functions in considered elements
  setRCInfo(linVCCSvec);
  setRCInfo(linfVCCSvec);
  setRCInfo(linVCQSvec);
  setRCInfo(genVCCSvec);
  setRCInfo(tdCSvec);
  setRCInfo(genVCQSvec);
  setRCInfo(indepCSvec);
  
  uptodate = true;
  return nterms;
}


void Circuit::print()
{
  int *iparm = options.getIHandler()->getParmVec();
  
  if (iparm[OPT_PRINTCIR]) {
    MacroElVector& mEvec = mVec.getVec();
    TerminalVector& termvec = tVec.getVec();
    
    assert(init_flag);
    printf("****************************************************************\n");
    printf("*                       Circuit Description                    *\n");
    printf("****************************************************************\n");
    printf("\n *** Circuit elements:\n\n");
    int imax = mEvec.size();
    for (int i(0); i < imax; i++) {
      ParmSet *pset = mEvec[i]->getParmSet();
      // Print element in "freeda" format. First type:name
      std::cout << pset->getInfo()->type << ":" << mEvec[i]->getName() << " ";
      int nterms = mEvec[i]->getNTerms();
      // Then terminals (nodes)
      for (int j(0); j < nterms; j++){
	std::cout << mEvec[i]->getTerminal(j)->getName() << " ";
      }
      // Then model name
      std::cout << " Model: " << pset->getName() << std::endl;
    }
    int cir_nterms = termvec.size();
    bool flag = false;
    for (int i(1); i < cir_nterms; i++) 
      if (!termvec[i]->getNConnections()) {
	if (!flag)
	  printf("\n *** Unconnected nodes:\n\n");
	std::cout << termvec[i]->getName() << std::endl;
	flag = true;
      }

    if (iparm[OPT_PRINTCIR] > 1) {
      printf("\n *** Device models:\n");
      ParmSetMap::iterator psetend = psetMap.end();
      for (ParmSetMap::iterator i(psetMap.begin()); i != psetend; i++) 
	i->second->print();
    }
    printf("\n");
  }
}


// Junk

// The following could one day go to a sanityCheck() routine:

//     // Look if this model is already in vector (time-consuming but
//     // unavoidable for now)
//     bool flag(false);
//     for (int j(0); j < npset; j++)
//       if (psetVec[j] == pset)
// 	flag = true;
//     if (!flag) {
//       psetVec[npset] = pset;
//       npset++;
//     }






