//
// Convergence seems to be reasonable but many iterations are usually
// required it the initial guess is bad.
// 
// by Carlos E. Christoffersen

#include <carrot/NewtonSolve.h>
#include <carrot/OP.h>
#include "global.h"

#include <cmath>

// Debug
//#define DEBUG 1

#ifdef DEBUG
#include <cstdio>
#endif

OP::OP(Circuit *cir)
{
  // Get the maximum number of controlling voltages we can expect
  const int& max_cv = cir->getMaxCV();
  // **********************  Work space vectors
  dij = new double[max_cv];
  vj = new double[max_cv];
  // Get the element vectors we need first
  genvccs = cir->getGenVCCSvec();
  linvccs = cir->getLinVCCSvec();
  indvccs = cir->getIndepCSvec();
  genvcqs = cir->getGenVCQSvec();
}

OP::~OP()
{
  delete [] vj;
  delete [] dij;
}

double OP::op(const int& nrcs, UMFPackDI& nam, 
	      double* uvec, double *svec,
	      int& iter, int& rejections, double* q)
{
  // ******************  Create NAM and source vectors

  // Erase source vector. 
  for (int i(-1); i<nrcs; i++)
    svec[i] = 0.;
  // Consider independent CS. This should be performed once since it
  // never changes (at least for a single DC point).
  for (IndepCSVector::iterator i = indvccs.begin(); i < indvccs.end(); i++) {
    // svec[-1] is allocated but not used anywhere
    double ivalue;
    // Think about time argument. Should we use t=0?
    // What about multi-time sources?
    (*i)->eval(ivalue, 0.);
    // Current flows from row 1 to 2.
    svec[(*i)->getRow1()] -= ivalue;
    svec[(*i)->getRow2()] += ivalue;
  }
  
  // Consider linear VCCSs: These never change (no time-varying linear
  // elements for now). We create this part of the matrix in triplet
  // form just once. Then we add/erase only the nonlinear part at each
  // iteration.
  for (LinVCCSVector::iterator i = linvccs.begin(); i < linvccs.end(); i++) {
    double g = (*i)->getGain();
    nam.addElement((*i)->getRow1(), (*i)->getCol1(), g);
    nam.addElement((*i)->getRow1(), (*i)->getCol2(), -g);
    nam.addElement((*i)->getRow2(), (*i)->getCol1(), -g);
    nam.addElement((*i)->getRow2(), (*i)->getCol2(), g);
  }
  // Mark the end of contribution of linear elements in matrix
  nam.setLinMark();

  // *********************************************************************
  // Solve nonlinear problem
  // *********************************************************************
  NewtonSolve<OP> nw(this, &nam, nrcs);
  double residual = nw.solve(uvec, svec, iter, rejections);

  // Save the charge values (if requested)
  if (q) {
    const int& nCharges = genvcqs.size();
    for (int i(0); i < nCharges; i++) {
      // Calculate vj.
      for (int k(0); k < genvcqs[i]->getNCVolt(); k++) {
	// uvec_j[-1] = 0 (takes care of ground node)
	vj[k] = uvec[genvcqs[i]->getColPlus()[k]]
	  - uvec[genvcqs[i]->getColMinus()[k]];
      }
      // Evaluate nonlinear element
      genvcqs[i]->eval(q[i], vj);
    }
  }

  return residual;
}

int OP::fillNAMandRHS(UMFPackDI& nam, double* svec, double* uvec_j, 
		       const double& beta, const double& gmin)
{
  int retcode = 1; // 1 means linear problem
  if (genvccs.size())
    retcode = 0; // 0 means normal nonlinear problem
  // Consider generic VCCSs
  for (unsigned i(0); i < genvccs.size(); i++) {
    // Calculate vj.
    for (int k(0); k < genvccs[i]->getNCVolt(); k++) {
      // uvec_j[-1] = 0 (takes care of ground node)
      vj[k] = uvec_j[genvccs[i]->getColPlus()[k]] 
	- uvec_j[genvccs[i]->getColMinus()[k]];
    }
    // Evaluate nonlinear element
    double ij;
    genvccs[i]->eval_and_deriv(ij, dij, vj);
    // Fill source vector. Current flows from row 1 to 2.
    svec[genvccs[i]->getRow1()] -= ij;
    svec[genvccs[i]->getRow2()] += ij;
    for (int k(0); k < genvccs[i]->getNCVolt(); k++) {
      // ************ TEST ************************
      // Scale up derivative to increase convergence rate
      dij[k] *= beta;
      // ************ end TEST ************************
      // svec[-1] is allocated but not used anywhere
      double ieq = dij[k] * vj[k];
      svec[genvccs[i]->getRow1()] += ieq;
      svec[genvccs[i]->getRow2()] -= ieq;
      // Fill NAM
      double geq = dij[k] + gmin;
      nam.addElement(genvccs[i]->getRow1(), genvccs[i]->getColPlus()[k], 
		     geq);
      nam.addElement(genvccs[i]->getRow1(), genvccs[i]->getColMinus()[k], 
		     -geq);
      nam.addElement(genvccs[i]->getRow2(), genvccs[i]->getColPlus()[k], 
		     -geq);
      nam.addElement(genvccs[i]->getRow2(), genvccs[i]->getColMinus()[k], 
		     geq);
    }
  }
  return retcode;
}



