#include <cassert>
#include <vector>
#include <carrot/UMFPackZI.h>

extern "C" {
#include <suitesparse/umfpack.h>
}

#include <cstdio>

UMFPackZI::~UMFPackZI()
{
  if (factor_flag)
    umfpack_zi_free_numeric(&Numeric);
  if (symbolic_flag) 
    umfpack_zi_free_symbolic(&Symbolic);
  if (flag) {
    delete [] W;
    delete [] Wi;

    delete [] colPtr;
    delete [] rowIdx;
    delete [] valA;
    delete [] Map;
  } 
}


void UMFPackZI::buildCompressed()
{
  int nz = rowvec.size();
  if (!flag) {
    // Allocate space for compressed col. format, generate matrix and
    // generate Map
    colPtr = new int[nrcs+1];
    rowIdx = new int[nz];
    valA = new double_complex[nz];
    Map = new int[nz];
    int status = umfpack_zi_triplet_to_col(nrcs, nrcs, nz, 
					   &(rowvec[0]), 
					   &(colvec[0]), 
					   (double*) &(valuevec[0]), NULL,
					   colPtr, rowIdx, 
					   (double*) valA, NULL, Map) ;
    assert(status == UMFPACK_OK);
    // Allocate space for workspace vectors
    Wi = new int[nrcs];
    W = new double[10 * nrcs];
    flag = true;
  }
  else {
    // colPtr and rowIdx should be the same. Recalculate valA from Map.
    for (int p(0) ; p < colPtr[nrcs] ; p++) 
      valA[p] = 0.;
    for (int k(0) ; k < nz ; k++) 
      valA[Map[k]] += valuevec[k] ;
  }
}

int UMFPackZI::symbolic()
{
  assert(flag);
  if (!symbolic_flag) {
    int status = umfpack_zi_symbolic(nrcs, nrcs, colPtr, rowIdx, 
				     (double*) valA, NULL, 
				     &Symbolic, NULL, NULL);
    //  assert(status == UMFPACK_OK);
    symbolic_flag = true;
    return status;
  } else 
    return 0;
}

int UMFPackZI::factor()
{
  assert(flag);
  int status = umfpack_zi_numeric(colPtr, rowIdx, (double*) valA, NULL,
				  Symbolic, &Numeric, NULL, NULL);
  // assert(status == UMFPACK_OK);
  factor_flag = true;
  return status;
}

int UMFPackZI::solve(double_complex* x, double_complex* rhs)
{
  assert(flag);
  int status = umfpack_zi_wsolve(UMFPACK_A, 
				 colPtr, rowIdx, (double*) valA, NULL,
				 (double*) x, NULL, (double*) rhs, NULL,
				 Numeric, NULL, NULL, Wi, W);
  // assert(status == UMFPACK_OK);
  return status;
}

int UMFPackZI::solveTransposed(double_complex* x, double_complex* rhs)
{
  assert(flag);
  int status = umfpack_zi_wsolve(UMFPACK_At, 
				 colPtr, rowIdx, (double*) valA, NULL,
				 (double*) x, NULL, (double*) rhs, NULL,
				 Numeric, NULL, NULL, Wi, W);
  // assert(status == UMFPACK_OK);
  return status;
}

void UMFPackZI::clearNonlinearNumeric()
{
  if (factor_flag) {
    umfpack_zi_free_numeric(&Numeric);
    // Erase only the nonlinear part of the triplet structure
    rowvec.resize(triplet_lin_size);
    colvec.resize(triplet_lin_size);
    valuevec.resize(triplet_lin_size);
    factor_flag = false;
  }
}

void UMFPackZI::clearCapNumeric()
{
  if (factor_flag) {
    umfpack_zi_free_numeric(&Numeric);
    // Erase the nonlinear part and capacitors of the triplet structure
    rowvec.resize(triplet_g_size);
    colvec.resize(triplet_g_size);
    valuevec.resize(triplet_g_size);
    factor_flag = false;
  }
}

void UMFPackZI::clearCompressed()
{
  if (flag) {
    delete [] colPtr;
    delete [] rowIdx;
    delete [] valA;
    delete [] Map;
    flag = false;
  } 
  if (symbolic_flag) { 
    umfpack_zi_free_symbolic(&Symbolic);
    symbolic_flag = false;
  }
}

void UMFPackZI::reset()
{
  // Erase the nonlinear part and capacitors of the triplet structure
  rowvec.resize(0);
  colvec.resize(0);
  valuevec.resize(0);
  if (factor_flag) {
    umfpack_zi_free_numeric(&Numeric);
    factor_flag = false;
  }
  if (flag) {
    delete [] colPtr;
    delete [] rowIdx;
    delete [] valA;
    delete [] Map;
    flag = false;
  } 
  if (symbolic_flag) { 
    umfpack_zi_free_symbolic(&Symbolic);
    symbolic_flag = false;
  }
}


void UMFPackZI::printTriplet()
{
  // This is useful for small matrices only
  if (nrcs < 31) {
    // Create a dense matrix
    int nc2 = nrcs * nrcs;
    double_complex *A = new double_complex[nc2];
    for (int i(0); i<nc2; i++)
      A[i] = 0.;
    for (unsigned i(0); i < rowvec.size(); i++)
      A[rowvec[i] + colvec[i] * nrcs] += valuevec[i];
    
    printf("\n");
    for (int i(0); i<nrcs; i++) {
      for (int j(0); j<nrcs; j++)
      if (A[i + nrcs*j] != 0.)
	printf("(%+.2e, ", A[i + nrcs*j].real());
      else
	printf("         0. ");
      printf("\n");
      for (int j(0); j<nrcs; j++)
      if (A[i + nrcs*j] != 0.)
	printf("%+.2e)  ", A[i + nrcs*j].imag());
      else
	printf("            ");
      printf("\n");
    }
    printf("\n");
  }
  else {
    printf("Matrix too large (%d), not printing\n", nrcs);
  }
}
