#include "../global.h"
#include "BJT.h"
#include <carrot/Options.h>
#include <carrot/FADBADGenericVCS.h> 
#include <carrot/FADBADTGenericVCS.h> 
#include <carrot/FADBADPowerSource.h> 
#include <carrot/FADBADLinearVCS.h> 

// Debug
//#include <cstdio>

static ItemInfo BJT_info = {
  "bjt",
  "NPN Bipolar Transistor",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

// ***********************************************************************
// Double parameters. We need a way to correlate this with the
// information in BJT_dparams
// ***********************************************************************
enum {IS, BF, NF, VAF, IKF, ISE, NE, BR, NR, VAR, IKR, ISC, NC,
      RE, RB, RBM, IRB, RC, EG, CJE, VJE, MJE,
      CJC, VJC, MJC, XCJC, FC, TF, XTF, VTF, ITF, TR,
      XTB, XTI, TNOM, TEMP, AREA, MAX_DPARM_INDEX}; 

static ParamHandler<double>::ParmInfo BJT_dparams[] = {
  {"is","Transport saturation current (A)", false, 1e-16},
  {"bf","Ideal maximum forward beta", false, 100.},
  {"nf","Forward current emission coefficient", false, 1.},
  {"vaf","Forward early voltage (V)", false, 0.},
  {"ikf","Forward-beta high current roll-off knee current (A)", false, 0.},
  {"ise","Base-emitter leakage saturation current (A)", false, 0.},
  {"ne","Base-emitter leakage emission coefficient", false, 1.5},
  {"br","Ideal maximum reverse beta", false, 1.},
  {"nr","Reverse current emission coefficient", false, 1.},
  {"var","Reverse early voltage (V)", false, 0.},
  {"ikr","Corner for reverse-beta high current roll off (A)", false, 0.},
  {"isc","Base collector leakage saturation current (A)", false, 0.},
  {"nc","Base-collector leakage emission coefficient", false, 2.},
  {"re","Emitter ohmic resistance (W)", false, 0.1},
  {"rb","Zero bias base resistance (W)", false, 0.1},
  {"rbm","Minimum base resistance (W)", false, 0.},
  {"irb","Current at which rb falls to half of rbm (A)", false, 0.},
  {"rc","Collector ohmic resistance (W)", false, 0.1},
  {"eg","Badgap voltage (eV)", false, 1.11},
  {"cje","Base emitter zero bias p-n capacitance (F)", false, 0.},
  {"vje","Base emitter built in potential (V)", false, 0.75},
  {"mje","Base emitter p-n grading factor", false, 0.33},
  {"cjc","Base collector zero bias p-n capacitance (F)", false, 0.},
  {"vjc","Base collector built in potential (V)", false, 0.75},
  {"mjc","Base collector p-n grading factor", false, 0.33},
  {"xcjc","Fraction of cbc connected internal to rb", false, 1.},
  {"fc","Forward bias depletion capacitor coefficient", false, 0.5},
  {"tf","Ideal forward transit time (S)", false, 0.},   
  {"xtf","Transit time bias dependence coefficient", false, 0.},
  {"vtf","Transit time dependency on vbc (V)", false, 0.},
  {"itf","Transit time dependency on ic (A)", false, 0.},
  {"tr","Ideal reverse transit time (S)", false, 0.},
  {"xtb","Forward and reverse beta temperature coefficient", false, 0.},
  {"xti","IS temperature effect exponent", false, 3.},
  {"tnom","Nominal temperature (C)", false, 27.},
  {"temp","Temperature (C). Defaults to global Temperature", false, 27.},
  {"area","Current multiplier", false, 1.}
};

// ***********************************************************************
// Integer parameters
// ***********************************************************************
enum {TYPE, TPORT, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo BJT_iparams[] = {
  {"type", "0 (NPN) or 1 (PNP)", false, 0},
  {"tport", "Use thermal port (5 terminals)", false, 0}
};

// ***********************************************************************
// Internal CS element definitions:
// ***********************************************************************
// This has to be included after enum (for double params) definition
#include "CommonParams.h"
#include "BJTBE.h"
#include "BJTBC.h"
#include "BJTCE.h"
#include "BJTRB.h"
#include "RCType.h"
#include "REType.h"
#include "RBType.h"
#include "QBE.h"
#include "QBC.h"
#include "QBX.h"

// #include "ThermalPowerS.h"
// template ThermalPowerS;

// *******************************
// Library loading mechanism
// *******************************
extern "C" {
  //Export the name of the module we are using
  const char * ModuleName(){
    return BJT_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		       &BJT_info, 
		       BJT_dparams, MAX_DPARM_INDEX,
		       BJT_iparams, MAX_IPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p=0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new BJT(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

// ***********************************************************************
// BJT members start:
// ***********************************************************************
BJT::BJT(ParmSet *pset, const std::string& name) 
  : MacroElement(pset, name)    
{
  // We need to define terminals here. Perhaps it was better to define
  // separate MacroElements for each type instead of this.
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  if (iparm[TPORT]) {
    setNTerms(5);
  }
  else 
    setNTerms(3);
}

BJT::~BJT()
{
  // There is not much to delete here because the circuit takes care
  // of deleting the CSs. (Is this right?)
}

void BJT::init()
{
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
  
  // This flow should be similar for many MacroElements:
  //
  // 0. Check parameters and terminals
  // (We should really throw and exception in the following:)
  getParmSet()->checkParms();
  assert(checkTerms());

  // Set default temperature if not specified in parameter set
  if (!(getParmSet()->getDHandler()->isSet(TEMP)))
    getParmSet()->getDHandler()->
      setParm(TEMP, options.getDHandler()->getParmVec()[OPT_TEMP]);

  if (iparm[TPORT]) {
    // Thigs to fix:
    // - Add power dissipation of linear resistors
    // - Improve device model support in general (i.e., fix terminal 
    //   vector hell, etc.)

    // internal resistors. Sorry, they can not be zero for now.
    assert(dparm[RC]);
    assert(dparm[RE]);
    assert(dparm[RB]);
    if (dparm[RBM]) {
      // Use a gyrator with the special VCCS
      setNInternalNodes(4);
      GenericVCCS* rb = new TBJTRB(dparm, NULL);
      GenericVCCS* psrb = new BJTRB_PS(dparm,NULL);
      if (!iparm[TYPE]) {
	// NPN
	static const int rbterms[10] = {-1,8,6,7,6,5,8,-1,3,4};
	static const int rbpsterms[10] = {4,3,6,7,6,5,8,-1,3,4};
	insert(rb, rbterms);
	insert(psrb, rbpsterms);
      } else {
	// PNP
	static const int rbterms[10] = {-1,8,7,6,5,6,8,-1,3,4};
	static const int rbpsterms[10] = {4,3,7,6,5,6,8,-1,3,4};
	insert(rb, rbterms);
	insert(psrb, rbpsterms);
      }
      // Gyrator sources (no need for parameter derivatives for these)
      LinearVCCS* g1 = new LinearVCCS(scale);
      static const int tvec_g1[4] = {8,-1,1,6};
      insert(g1, tvec_g1);
      LinearVCCS* g2 = new LinearVCCS(scale);
      static const int tvec_g2[4] = {1,6,8,-1};
      insert(g2, tvec_g2);
    }
    else {
      // Use a regular resistor
      setNInternalNodes(3);
      LinearVCCS* rb = new RBType(dparm, NULL);
      static const int rbterms[4] = {6,1,6,1};
      insert(rb,rbterms);
    }

    LinearVCCS* rc = new RCType(dparm, NULL);
    static const int rcterms[4] = {5,0,5,0};
    insert(rc, rcterms);
    LinearVCCS* re = new REType(dparm, NULL);
    static const int reterms[4] = {7,2,7,2};
    insert(re, reterms);

    // Nonlinear elements
    // Intrinsic model
    GenericVCCS* ibe = new TBJTBE(dparm, NULL);
    GenericVCCS* ibc = new TBJTBC(dparm, NULL);
    GenericVCCS* ice = new TBJTCE(dparm, NULL);
    GenericVCCS* psibe = new BJTBE_PS(dparm, NULL);
    GenericVCCS* psibc = new BJTBC_PS(dparm, NULL);
    GenericVCCS* psice = new BJTCE_PS(dparm, NULL);
    if (!iparm[TYPE]) {
      // NPN transistor
      static const int ibeterms[6] = {6,7,6,7,3,4};
      static const int ibepsterms[6] = {4,3,6,7,3,4};
      insert(ibe, ibeterms);
      insert(psibe, ibepsterms);
      static const int ibcterms[6] = {6,5,6,5,3,4};      
      static const int ibcpsterms[6] = {4,3,6,5,3,4};      
      insert(ibc, ibcterms);
      insert(psibc, ibcpsterms);
      static const int iceterms[8] = {5,7,6,7,6,5,3,4};
      static const int icepsterms[8] = {4,3,6,7,6,5,3,4};
      insert(ice, iceterms);
      insert(psice, icepsterms);
      // Capacitors
      if (dparm[TF] || dparm[CJE]) {
	GenericVCQS* qbe = new TQBE(dparm, NULL);
	static const int qbeterms[8] = {6,7,6,7,6,5,3,4};	
	insert(qbe, qbeterms);
      }
      if (dparm[TR] || dparm[CJC]) {
	GenericVCQS* qbc = new TQBC(dparm, NULL);
	insert(qbc, ibcterms);
	if (dparm[XCJC] < 1.) {
	  GenericVCQS* qbx = new TQBX(dparm, NULL);
	  static const int qbxterms[6] = {1,5,1,5,3,4};
	  insert(qbx, qbxterms);
	}
      }   
    } else {
      // PNP transistor
      // Just invert currents and voltages
      static const int ibeterms[6] = {7,6,7,6,3,4};
      static const int ibepsterms[6] = {4,3,7,6,3,4};
      insert(ibe, ibeterms);
      insert(psibe, ibepsterms);
      static const int ibcterms[6] = {5,6,5,6,3,4};      
      static const int ibcpsterms[6] = {4,3,5,6,3,4};      
      insert(ibc, ibcterms);
      insert(psibc, ibcpsterms);
      static const int iceterms[8] = {7,5,7,6,5,6,3,4};
      static const int icepsterms[8] = {4,3,7,6,5,6,3,4};
      insert(ice, iceterms);
      insert(psice, icepsterms);
      // Capacitors
      if (dparm[TF] || dparm[CJE]) {
	GenericVCQS* qbe = new TQBE(dparm, NULL);
	static const int qbeterms[8] = {7,6,7,6,5,6,3,4};	
	insert(qbe, qbeterms);
      }
      if (dparm[TR] || dparm[CJC]) {
	GenericVCQS* qbc = new TQBC(dparm, NULL);
	insert(qbc, ibcterms);
	if (dparm[XCJC] < 1.) {
	  GenericVCQS* qbx = new TQBX(dparm, NULL);
	  static const int qbxterms[6] = {5,1,5,1,3,4};
	  insert(qbx, qbxterms);
	}
      }   
    }
  }
  else {
    // internal resistors. Sorry, they can not be zero for now.
    assert(dparm[RC]);
    assert(dparm[RE]);
    assert(dparm[RB]);
    if (dparm[RBM]) {
      // Use a gyrator with the special VCCS
      setNInternalNodes(4);
      GenericVCCS* rb = new BJTRB(dparm, NULL);
      if (!iparm[TYPE]) {
	// NPN
	static const int rbterms[8] = {-1,6,4,5,4,3,6,-1};
	insert(rb, rbterms);
      } else {
	// PNP
	static const int rbterms[8] = {-1,6,5,4,3,4,6,-1};
	insert(rb, rbterms);
      }
      // Gyrator sources (no need for parameter derivatives for these)
      LinearVCCS* g1 = new LinearVCCS(scale);
      static const int tvec_g1[4] = {6,-1,1,4};
      insert(g1, tvec_g1);
      LinearVCCS* g2 = new LinearVCCS(scale);
      static const int tvec_g2[4] = {1,4,6,-1};
      insert(g2, tvec_g2);
    }
    else {
      // Use a regular resistor
      setNInternalNodes(3);
      LinearVCCS* rb = new RBType(dparm, NULL);
      static const int rbterms[4] = {4,1,4,1};
      insert(rb,rbterms);
    }

    LinearVCCS* rc = new RCType(dparm, NULL);
    static const int rcterms[4] = {3,0,3,0};
    insert(rc, rcterms);
    LinearVCCS* re = new REType(dparm, NULL);
    static const int reterms[4] = {5,2,5,2};
    insert(re, reterms);

    // Nonlinear elements
    // Intrinsic model
    GenericVCCS* ibe = new BJTBE(dparm, NULL);
    GenericVCCS* ibc = new BJTBC(dparm, NULL);
    GenericVCCS* ice = new BJTCE(dparm, NULL);
    if (!iparm[TYPE]) {
      // NPN transistor
      static const int ibeterms[4] = {4,5,4,5};
      insert(ibe, ibeterms);
      static const int ibcterms[4] = {4,3,4,3};      
      insert(ibc, ibcterms);
      static const int iceterms[6] = {3,5,4,5,4,3};
      insert(ice, iceterms);
      // Capacitors
      if (dparm[TF] || dparm[CJE]) {
	GenericVCQS* qbe = new QBE(dparm, NULL);
	static const int qbeterms[6] = {4,5,4,5,4,3};	
	insert(qbe, qbeterms);
      }
      if (dparm[TR] || dparm[CJC]) {
	GenericVCQS* qbc = new QBC(dparm, NULL);
	insert(qbc, ibcterms);
	if (dparm[XCJC] < 1.) {
	  GenericVCQS* qbx = new QBX(dparm, NULL);
	  static const int qbxterms[4] = {1,3,1,3};
	  insert(qbx, qbxterms);
	}
      }   
    } else {
      // PNP transistor
      // Just invert currents and voltages
      static const int ibeterms[4] = {5,4,5,4};
      insert(ibe, ibeterms);
      static const int ibcterms[4] = {3,4,3,4};      
      insert(ibc, ibcterms);
      static const int iceterms[6] = {5,3,5,4,3,4};
      insert(ice, iceterms);
      // Capacitors
      if (dparm[TF] || dparm[CJE]) {
	GenericVCQS* qbe = new QBE(dparm, NULL);
	static const int qbeterms[6] = {5,4,5,4,3,4};	
	insert(qbe, qbeterms);
      }
      if (dparm[TR] || dparm[CJC]) {
	GenericVCQS* qbc = new QBC(dparm, NULL);
	insert(qbc, ibcterms);
	if (dparm[XCJC] < 1.) {
	  GenericVCQS* qbx = new QBX(dparm, NULL);
	  static const int qbxterms[4] = {3,1,3,1};
	  insert(qbx, qbxterms);
	}
      }   
    }

  }
}


