// This may look like C code, but it is really -*- C++ -*-
//
// Implements the effect of RBM in the bipolar transistor. Must be
// applied though a gyrator (because we can only evaluate ib*rb)
//
// 

#ifndef BJTRB_H
#define BJTRB_H 1

class BJTRB_eval {
public:
  // This are locally-defined variables
  enum {T_IS, VT, MAX_TVAR_INDEX};
  enum {C_K1, C_K2, MAX_CVAR_INDEX};

  BJTRB_eval(int *iparm) {} // iparm not needed here

  // Main evaluation function. Only voltage dependence considered here.
  template <typename currDouble, typename voltDouble, typename paramDouble, 
	    typename varDouble, typename cvarDouble>
  inline void operator()(currDouble& vrb, voltDouble* v, 
			 paramDouble* dparm, 
			 varDouble* tvar, cvarDouble* cvar)
  {
    // Limit the maximum value of the exponential function
    // (consider a special exp() function).
    // v[0] : Vbe
    // v[1] : Vbc
    // v[2] : ib / SCALE
    const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
    currDouble rb;
    currDouble ib = v[2] * scale;
    if (dparm[IRB] == 0.) {
      // Use kqb
      currDouble ibf;
      getIexp(ibf, v[0], tvar[T_IS], dparm[NF], tvar[VT]);
      currDouble ibr;
      getIexp(ibr, v[1], tvar[T_IS], dparm[NR], tvar[VT]);
      currDouble kqb;
      getKqb(kqb, v[0], v[1], ibf, ibr, dparm, tvar[T_IS], tvar[VT]);
      rb = dparm[RBM] + (dparm[RB] - dparm[RBM]) / kqb / dparm[AREA];
    }
    else {
      // Use IRB
      if (v[2] != 0.) {
	// Make sure ib1 is always positive
	currDouble ib1;
	if (ib > 0.) 
	  ib1 = ib;
	else
	  ib1 = -ib;
	currDouble x = sqrt(1.+ cvar[C_K1]*ib1) - 1.;
	x *= cvar[C_K2] / sqrt(ib1);
	currDouble tx = tan(x);
	rb = dparm[RBM] + 3. * (dparm[RB] - dparm[RBM]) 
	  * (tx - x) / (x * tx * tx) / dparm[AREA];
      } else 
	rb = dparm[RB] / dparm[AREA];
    }
    vrb = rb * ib * scale;
  }

  // Adjust variables to a new temperature.
  template <typename tvarDouble, typename varDouble, typename tempDouble>
  inline void setTemp(tvarDouble* tvar, 
		      varDouble* dparm, varDouble* cvar, tempDouble& T)
  {
    tempDouble Tabs = T + tzero;
    // Parameter calculation 
    getVt(tvar[VT], Tabs);
    getIs(tvar[T_IS], dparm[IS], dparm[NF], tvar[VT], dparm, Tabs);
  }
  
  // For efficiency, we may want to pre-calculate some expressions
  // that do not change with voltage or temperature.  Adjust variables
  // to a new parameter set.
  template <typename varDouble>
  inline void newParms(varDouble* cvar, varDouble* dparm)
  {
    cvar[C_K1] = 144. / dparm[IRB] / dparm[AREA] /pi/pi;
    cvar[C_K2] = pi*pi * sqrt(dparm[IRB] * dparm[AREA]) / 24.;
  }

  // Calculate power. pout initially contains the current.
  template <typename powerDouble, typename voltDouble, 
	    typename paramDouble>
  inline void power(powerDouble& pout, voltDouble* v, 
		    paramDouble* dparm)
  {
    pout *= v[2]; // scale cancels out
  }

  //private:
  
  // Pointer to integer parameter vector from macroelement
  //  int *iparm; // not necessary for this model

};

// No thermal port
typedef FADBADGenericVCS<BJTRB_eval, GenericVCCS, 3, 
			 MAX_DPARM_INDEX, TEMP, false> BJTRB;
// With thermal port
typedef FADBADTGenericVCS<BJTRB_eval, GenericVCCS, 4, 
			  MAX_DPARM_INDEX, TEMP> TBJTRB;
// Power source
typedef FADBADPowerSource<BJTRB_eval, 4, 
 			  MAX_DPARM_INDEX, TEMP> BJTRB_PS;


#endif
