#include "global.h"
#include "MESFET.h"
#include <carrot/Options.h>
#include <carrot/FADBADGenericVCS.h> 
#include <carrot/FADBADTGenericVCS.h> 
#include <carrot/FADBADLinearVCS.h> 
#include <carrot/FADBADPowerSource.h> 

static ItemInfo MESFET_info = {
  "mesfet",
  "GaAs MESFET and HEMT",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

// Double Parameters
// Make sure enum and ParInfo definitions are in the same order!
enum {A0, A1, A2, A3, BETA, VDS0, GAMA, VT0, CGS0, CGD0, IS, N, IB0, NR, 
      T, RI, VBI, FCC, VBD, TNOM, TEMP, AVT0, BVT0, TBET, TM, TME, EG, 
      MGS, MGD, XTI, AREA, MAX_DPARM_INDEX};

//Insert ParmInfo values here
static ParamHandler<double>::ParmInfo MESFET_dparams[] = { 
  {"a0", "Drain saturation current for Vgs=0 (A)", false, .1},
  {"a1", "Coefficient for V1 (A/V)", false, .05},
  {"a2", "Coefficient for V1^2 (A/V^2)", false, 0.},
  {"a3", "Coefficient for V1^3 (A/V^3)", false, 0.},
  {"beta", "V1 dependance on Vds (1/V)", false, 0.},
  {"vds0", "Vds at which BETA was measured (V)", false, 4.},
  {"gama", "Slope of drain characteristic in the linear region (1/V)", 
   false, 1.5},
  {"vt0", 
   "Voltage at which the channel current is forced to be zero for Vgs<=Vto (V)", 
   false, -1e10},
  {"cgs0", "Gate-source Schottky barrier capacitance for Vgs=0 (F)",
   false, 0.},
  {"cgd0", "Gate-drain Schottky barrier capacitance for Vgd=0 (F)",
   false, 0.},
  {"is", "Diode saturation current (A)", false, 0.},
  {"n", "Diode ideality factor", false, 1.},
  {"ib0", "Breakdown current parameter (A)", false, 0.},
  {"nr", "Breakdown ideality factor", false, 10.},
  {"t", "Channel transit time *not implemented yet* (s)", false, 0.},
  {"ri", "Channel resistance (Ohm)", false, 0.},
  {"vbi", "Built-in potential of the Schottky junctions (V)", false, .8},
  {"fcc", "Forward-bias depletion capacitance coefficient (V)", false, .5},
  {"vbd", "Breakdown voltage (V)", false, 1e10},
  {"tnom", "Nominal temperature (C)", false, 27.},
  {"temp", "Temperature (C). Defaults to global temp.", false, 27.},
  {"avt0", "Pinch-off voltage (VP0 or VT0) linear temp. coefficient (1/K)",
   false, 0.},
  {"bvt0", 
   "Pinch-off voltage (VP0 or VT0) quadratic temp. coefficient (1/K^2)",
   false, 0.},
  {"tbet", "BETA power law temperature coefficient (1/K)", false, 0.},
  {"tm", "Ids linear temp. coeff. (1/K)", false, 0.},
  {"tme", "Ids power law temp. coeff. (1/K^2)", false, 0.},
  {"eg", "Barrier height at 0.K (eV)", false, .8},
  {"mgs", "Gate-Source grading coefficient", false, .5},
  {"mgd", "Gate-Drain grading coefficient", false, .5},
  {"xti", "Diode saturation current temperature exponent", false, 2.},
  {"area", "Area multiplier", false, 1.}
};

// Uncomment for integer parameters
enum {TPORT, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo MESFET_iparams[] = {
//   {"type", "Model type: 1-Curtice Cubic", false, 1},
  {"tport", "Use thermal port (5 terminals)", false, 0}
};

//Library loading mechanism
extern "C" {
  const char * ModuleName(){
    return MESFET_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		       &MESFET_info,
		       MESFET_dparams, MAX_DPARM_INDEX, 
		       MESFET_iparams, MAX_IPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p=0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new MESFET(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

// Include current source models
#include "CommonCurtice.h"
#include "IdsCurtice.h"
#include "IgsCurtice.h"
#include "IgdCurtice.h"
#include "QgsCurtice.h"
#include "QgdCurtice.h"
#include "RIType.h"


MESFET::MESFET(ParmSet *pset, const std::string& name)
  :MacroElement(pset, name)
{
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  if (iparm[TPORT])
    setNTerms(5);
  else 
    setNTerms(3);
  // Just try this to make sure no enums are missing
  assert(MESFET_dparams[AREA].defaultValue == 1.);
}

MESFET::~MESFET()
{
  //Destructor code goes here
}

void MESFET::init()
{
  // access to double and int parameters
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();

  // This flow should be similar for many MacroElements:
  //
  // 0. Check parameters and terminals
  // (We should really throw and exception in the following:)
  // Check required parameters
  getParmSet()->checkParms();
  // Check external terminal connections
  assert(checkTerms());

  // Set default temperature if not specified in parameter set
  if (!(getParmSet()->getDHandler()->isSet(TEMP)))
    getParmSet()->getDHandler()->
      setParm(TEMP, options.getDHandler()->getParmVec()[OPT_TEMP]);
  // Assign connection vector
  static const int igsterms[6] = {1,2,1,2,3,4};
  static const int igdterms[6] = {1,0,1,0,3,4};
  static const int idsterms[8] = {0,2,1,2,0,2,3,4};
  GenericVCCS* igs;
  GenericVCCS* igd;
  GenericVCCS* ids;
  // Pointer for cgs terminal vector
  const int* cgsptr;
  // If RI is specified, create internal node
  if (dparm[RI] > 0.) {
    setNInternalNodes(1);
    LinearVCCS* ri = new RIType(dparm, NULL);
    const int* riptr;
    if (!iparm[TPORT]) {
      static const int cgsterms[4] = {1,3,1,3};
      cgsptr = cgsterms;
      static const int riterms[4] = {3,2,3,2};
      riptr = riterms;
    } else {
      // Assign connection vector
      static const int cgsterms[6] = {1,5,1,5,3,4};
      cgsptr = cgsterms;
      static const int riterms[4] = {5,2,5,2};
      riptr = riterms;
    }
    insert(ri, riptr);
  } else {
    setNInternalNodes(0);
    cgsptr = igsterms;
  }

  if (!iparm[TPORT]) {
    // Nonlinear elements
    igs = new IgsCurtice(dparm, NULL);
    igd = new IgdCurtice(dparm, NULL);
    ids = new IdsCurtice(dparm, NULL);
  } else {
    // Nonlinear elements
    igs = new TIgsCurtice(dparm, NULL);
    igd = new TIgdCurtice(dparm, NULL);
    ids = new TIdsCurtice(dparm, NULL);
    // Power generators
    GenericVCCS* pgs = new IgsPS(dparm, NULL);
    GenericVCCS* pgd = new IgdPS(dparm, NULL);
    GenericVCCS* pds = new IdsPS(dparm, NULL);
    static const int pgsterms[6] = {4,3,1,2,3,4};
    static const int pgdterms[6] = {4,3,1,0,3,4};
    static const int pdsterms[8] = {4,3,1,2,0,2,3,4};
    insert(pgs, pgsterms);
    insert(pgd, pgdterms);
    insert(pds, pdsterms);
  }
  insert(igs, igsterms);
  insert(igd, igdterms);
  insert(ids, idsterms);
  // Capacitors
  if (dparm[CGD0]) {
    GenericVCQS* qgd = new QgdCurtice(dparm, NULL);
    insert(qgd, igdterms);
  }
  if (dparm[CGS0]) {
    GenericVCQS* qgs = new QgsCurtice(dparm, NULL);
    insert(qgs, cgsptr);
  }
}

