#include "global.h"
#include "MVsin.h"
#include <carrot/Options.h>
#include <carrot/FADBADMIndependentCS.h> 
#include <carrot/FADBADLinearVCS.h> 

static ItemInfo MVsin_info = {
  "mvsin",
  "Multi-time sinusoidal voltage source",
  "Carlos Christoffersen"
  DEFAULT_ADDRESS
};

// Double Parameters
// Make sure enum and ParInfo definitions are in the same order!
enum {AMP, FREQ, PHASE, RI, TR, AMT, AMF, CARRIER, MAX_DPARM_INDEX};

//Insert ParmInfo values here
static ParamHandler<double>::ParmInfo MVsin_dparams[] = { 
  {"amp", "Amplitude (V)", true, 0.},
  {"freq", "Frequency in t2 scale (Hz)", true, 0.},
  {"phase", "Phase (rad)", false, 0.},
  {"ri", "Internal resistance (Ohms)", false, 0.},
  {"tr", "Rise time in t1 (s)", false, 0.},
  {"am_t", "AM period (s)", false, 0.},
  {"am_f", "AM sin() factor", false, 0.},
  {"carrier", "Normalized carrier amplitude", false, 0.}
};

// Uncomment for integer parameters
//enum {MAX_IPARM_INDEX};
//static ParamHandler<int>::ParmInfo MVsin_iparams[] = {  };


//Library loading mechanism
extern "C" {
  const char * ModuleName(){
    return MVsin_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		         &MVsin_info,
		         MVsin_dparams, MAX_DPARM_INDEX); 
//		         MVsin_iparams, MAX_IPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p=0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new MVsin(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

// ***********************************************************************
// Internal CS element definitions:
// ***********************************************************************
// This has to be included after enum (for double params) definition
#include "InternalMSinCS.h"
#include "RIType.h"

MVsin::MVsin(ParmSet *pset, const std::string& name)
  :MacroElement(pset, name, 2)
{
  //Constructor code goes here (above, 2 is the number of terminals)
}

MVsin::~MVsin()
{
  //Destructor code goes here
}

void MVsin::init()
{
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
  // This flow should be similar for many MacroElements:
  //
  // 0. Check parameters and terminals
  // (We should really throw and exception in the following:)
  getParmSet()->checkParms();
  assert(checkTerms());
  // 1. Create individual CSs
  InternalMSinCS* dccs = new InternalMSinCS(dparm, NULL);
  if (dparm[RI] == 0.) {
    // Need gyrator: topology 1
    // "ri" parameter not used in this case. Perhaps a warning would
    // be appropriate here if ri was set.
    assert(scale != 0.);
    // Create internal terminal (voltage of this is current of voltage source)
    // We also need a ground terminal (not counted as internal)
    setNInternalNodes(1);
    static const int tvec_source[2] = {-1,2};
    insert(dccs, tvec_source);

    // Gyrator sources
    LinearVCCS* g1 = new LinearVCCS(scale);
    static const int tvec_g1[4] = {2,-1,0,1};
    insert(g1, tvec_g1);
    LinearVCCS* g2 = new LinearVCCS(scale);
    static const int tvec_g2[4] = {0,1,2,-1};
    insert(g2, tvec_g2);
  } else {
    // No internal nodes necessary
    setNInternalNodes(0);
    static const int tvec_source[2] = {1,0};
    insert(dccs, tvec_source);

    // No gyrator is necessary: topology 2
    LinearVCCS* gi = new RIType(dparm, NULL);
    static const int tvec_gi[4] = {0,1,0,1};
    insert(gi, tvec_gi);
  }
}



