// This may look like C code, but it is really -*- C++ -*-
//
// Implements the charge stored in a diode
//
//             
//          o  1              
//          |                            
//        --+--
//               DiodeCharge  
//        --+--
//          |                          
//          o  0    	      
//                
// 

#ifndef DiodeCharge_H
#define DiodeCharge_H 1

class DiodeCharge_eval {
public:
  // This are locally-defined variables
  enum {T_IS, VT, T_EGAP, T_VJ, T_CJ0, k5, k6, k7, MAX_TVAR_INDEX};
  enum {k1, k2, k3, k4, EGAPn, MAX_CVAR_INDEX};

  DiodeCharge_eval(int *iparm) {} // iparm not needed here

  // Main evaluation function. Only voltage dependence considered here.
  template <typename currDouble, typename voltDouble, typename paramDouble, 
	    typename varDouble, typename cvarDouble>
  inline void operator()(currDouble& qd, voltDouble* v, 
			 paramDouble* dparm, 
			 varDouble* tvar, cvarDouble* cvar)
  {
    if (dparm[CJ0] > 0.) {
      if (v[0] < dparm[FC] * tvar[T_VJ]) {
	qd = tvar[k5] * (1. - pow(1. - v[0] / tvar[T_VJ], cvar[k4]));
      } else {
	qd = tvar[k6] * ((1. - dparm[FC] * (1. + dparm[M])) * v[0] 
			 + .5 * dparm[M] * v[0] * v[0] / tvar[T_VJ] 
			 - tvar[k7])
	  + tvar[k5] * (1. - pow(1. - dparm[FC], cvar[k4]));
      }
    } else 
      qd = 0.;
    if (dparm[TT] > 0.) {
      // Need to calculate id for TT component
      currDouble id = dparm[AREA] * tvar[T_IS] 
	* (safe_exp(v[0] / dparm[N] / tvar[VT]) - 1.);
      qd += dparm[TT] * id;
    }
    // Area effect
    qd *= dparm[AREA];
  }

  // Adjust variables to a new temperature.
  template <typename tvarDouble, typename varDouble, typename tempDouble>
  inline void setTemp(tvarDouble* tvar, 
		      varDouble* dparm, varDouble* cvar, tempDouble& T)
  {
    tempDouble Tabs = T + tzero;
    // Parameter calculation 
    tvarDouble tn = Tabs / (dparm[TNOM]+tzero);
    tvar[T_IS] = dparm[IS] * pow(tn, cvar[k1]) 
      * exp(cvar[k2] * Tabs - cvar[k3]); 
    tvar[VT] = kBoltzman * Tabs / eCharge;
    tvar[T_EGAP] = dparm[EG0] - .000702 * Tabs * Tabs / (Tabs + 1108.);
    tvar[T_VJ] = dparm[VJ] * tn - 3. * tvar[VT] * log(tn) 
      - tn * cvar[EGAPn] + tvar[T_EGAP];
    tvar[T_CJ0] = dparm[CJ0]
      * (1. + dparm[M] * (.0004 * (T - dparm[TNOM]) 
    			  + 1. - tvar[T_VJ] / dparm[VJ]));
    tvar[k5] = tvar[T_VJ] * tvar[T_CJ0] / cvar[k4];
    tvar[k6] = tvar[T_CJ0] * pow(1. - dparm[FC], - dparm[M] - 1.);
    tvar[k7] = ((1. - dparm[FC] * (1. + dparm[M])) * tvar[T_VJ] * dparm[FC]
		+ .5 * dparm[M] * tvar[T_VJ] * dparm[FC] * dparm[FC]);
  }
  
  // For efficiency, we may want to pre-calculate some expressions
  // that do not change with voltage nor temperature.  Adjust variables
  // to a new parameer set.
  template <typename varDouble>
  inline void newParms(varDouble* cvar, varDouble* dparm)
  {
    cvar[k1] = dparm[XTI]/dparm[N];
    cvar[k2] = eCharge * dparm[EG0] / dparm[N] / kBoltzman 
      / (dparm[TNOM] + tzero);
    cvar[k3] = eCharge * dparm[EG0] / dparm[N] / kBoltzman;
    cvar[k4] = 1. - dparm[M];
    varDouble TnomAbs = dparm[TNOM] + tzero;
    cvar[EGAPn] = dparm[EG0] - .000702 * TnomAbs * TnomAbs 
      / (TnomAbs + 1108.);
  }

  //private:
  
  // Pointer to integer parameter vector from macroelement
  //  int *iparm; // not necessary for this model

};


typedef FADBADGenericVCS<DiodeCharge_eval, GenericVCQS, 1, 
			 MAX_DPARM_INDEX, TEMP, false> DiodeCharge;
// With thermal port
typedef FADBADTGenericVCS<DiodeCharge_eval, GenericVCQS, 2, 
			  MAX_DPARM_INDEX, TEMP> TDiodeCharge;

#endif
