// This may look like C code, but it is really -*- C++ -*-
//
// Implements a the intrinsic diode junction (no cap) with tempereture effects.
//
//             
//          o  1                 o  2
//          |                            
//        --+--
//         / \   IntrinsicDiode  (temperature port in TIntrinsicDiode)
//        '-+-'
//          |                          
//          o  0    	         o  3     
//                
// 

#ifndef IntrinsicDiode_H
#define IntrinsicDiode_H 1

class IntrinsicDiode_eval {
public:
  // These are locally-defined variables
  enum {T_IS, VT, MAX_TVAR_INDEX};
  enum {k1, k2, k3, MAX_CVAR_INDEX};

  IntrinsicDiode_eval(int *iparm) {} // iparm not needed here

  // Main evaluation function. Only voltage dependence considered here.
  template <typename currDouble, typename voltDouble, typename paramDouble, 
	    typename varDouble, typename cvarDouble>
  inline void operator()(currDouble& id, voltDouble* v, 
			 paramDouble* dparm, 
			 varDouble* tvar, cvarDouble* cvar)
  {
    // Limit the maximum value of the exponential function
    // (consider a special exp() function).
    id = tvar[T_IS] * (safe_exp(v[0] / dparm[N] / tvar[VT]) - 1.);
    if (dparm[BV] > 0.)
      // Add breakdown current
      id -= dparm[IBV] * safe_exp(-(v[0] + dparm[BV]) / dparm[N] / tvar[VT]);
    // Area effect
    id *= dparm[AREA];
  }

  // Adjust variables to a new temperature.
  template <typename tvarDouble, typename varDouble, typename tempDouble>
  inline void setTemp(tvarDouble* tvar, 
		      varDouble* dparm, varDouble* cvar, tempDouble& T)
  {
    // Parameter calculation 
    tvar[T_IS] = dparm[IS] * pow((T+tzero)/(dparm[TNOM]+tzero), cvar[k1]) 
      * exp(cvar[k2] - cvar[k3] / T); 
    tvar[VT] = kBoltzman * (T + tzero) / eCharge;
  }
  
  // For efficiency, we may want to pre-calculate some expressions
  // that do not change with voltage or temperature.  Adjust variables
  // to a new parameer set.
  template <typename varDouble>
  inline void newParms(varDouble* cvar, varDouble* dparm)
  {
    cvar[k1] = dparm[XTI]/dparm[N];
    cvar[k2] = eCharge * dparm[EG0] / dparm[N] / kBoltzman 
      / (dparm[TNOM]+tzero);
    cvar[k3] = eCharge * dparm[EG0] / dparm[N] / kBoltzman;
  }

  // Calculate power. pout initially contains the current.
  template <typename powerDouble, typename voltDouble, 
	    typename paramDouble>
  inline void power(powerDouble& pout, voltDouble* v, 
		    paramDouble* dparm)
  {
    pout *= v[0];
  }

  //private:
  
  // Pointer to integer parameter vector from macroelement
  //  int *iparm; // not necessary for this model

};

// No thermal port
typedef FADBADGenericVCS<IntrinsicDiode_eval, GenericVCCS, 1, 
			 MAX_DPARM_INDEX, TEMP, false> IntrinsicDiode;
// With thermal port
typedef FADBADTGenericVCS<IntrinsicDiode_eval, GenericVCCS, 2, 
			  MAX_DPARM_INDEX, TEMP> TIntrinsicDiode;
// Power source
typedef FADBADPowerSource<IntrinsicDiode_eval, 2, 
 			  MAX_DPARM_INDEX, TEMP> PowerSource;


#endif
