#include "global.h"
#include "SpiceDiode.h"
#include <carrot/FADBADGenericVCS.h> 
#include <carrot/FADBADTGenericVCS.h> 
#include <carrot/FADBADPowerSource.h> 
#include <carrot/FADBADLinearVCS.h> 
#include <carrot/Options.h>
// Debug
//#include <cstdio>

static ItemInfo SpiceDiode_info = {
  "diode",
  "Spice Diode",
  "Carlos Christoffersen",
  DEFAULT_ADDRESS
};

// ***********************************************************************
// Double parameters. We need a way to correlate this with the
// information in SpiceDiode_dparams
// ***********************************************************************
enum {IS, N, IBV, BV, FC, CJ0, VJ, M, TT, AREA, RS, XTI, EG0, 
      TNOM, TEMP, MAX_DPARM_INDEX}; 

static ParamHandler<double>::ParmInfo SpiceDiode_dparams[] = {
  {"is", "Saturation current (A)", false, 1e-14},
  {"n", "Emission coefficient", false, 1.},
  {"ibv", "Current magnitude at the reverse breakdown voltage (A)", 
   false, 1e-10},
  {"bv", "Breakdown voltage (V)", false, 0.},
  {"fc", "Coefficient for forward-bias depletion capacitance", 
   false, .5},
  {"cj0", "Zero-bias depletion capacitance (F)", false, 0.},
  {"vj", "Built-in junction potential (V)", false, 1.},
  {"m", "PN junction grading coefficient", false, .5},
  {"tt", "Transit time (s)", false, 0.},
  {"area", "Area multiplier", false, 1.},
  {"rs", "Series resistance (ohms)", false, 0.},
  {"xti", "Is temperature exponent", false, 3.},
  {"eg0", "Energy bandgap (eV)", false, 1.11},
  {"tnom", "Nominal temperature (C)", false, 27.},
  {"temp", "Temperature (C). Defaults to global Temperature", 
   false, 27.}
};

// ***********************************************************************
// Integer parameters
// ***********************************************************************
enum {TPORT, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo SpiceDiode_iparams[] = {
  {"tport", "Use thermal port (4 terminals)", false, 0}
};

extern "C" {

  //Export the name of the module we are using
  const char * ModuleName(){
    return SpiceDiode_info.type.c_str();
  }

  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		       &SpiceDiode_info, 
		       SpiceDiode_dparams, MAX_DPARM_INDEX,
		       SpiceDiode_iparams, MAX_IPARM_INDEX);
  }

  void delParmSet(ParmSet * p){
    delete p;
    p = 0;
  }

  MacroElement * newObject(ParmSet * p, const char * name){
    return new SpiceDiode(p, name);
  }

  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

// ***********************************************************************
// Internal CS element definitions:
// ***********************************************************************
// This has to be included after enum (for double params) definition
#include "IntrinsicDiode.h"
#include "DiodeCharge.h"
#include "RSType.h"

//#include <carrot/ThermalPowerS.h>
//template ThermalPowerS;

// ***********************************************************************
// SpiceDiode members start:
// ***********************************************************************
SpiceDiode::SpiceDiode(ParmSet *pset, const std::string& name) 
  : MacroElement(pset, name)    
{
  // We need to define terminals here. Perhaps it was better to define
  // separate MacroElements for each type instead of this.
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  if (iparm[TPORT])
    setNTerms(4);
  else 
    setNTerms(2);
}

SpiceDiode::~SpiceDiode()
{
  // There is not much to delete here because the circuit takes care
  // of deleting the CSs. (Is this right?)
}

void SpiceDiode::init()
{
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  // This flow should be similar for many MacroElements:
  //
  // 0. Check parameters and terminals
  // (We should really throw and exception in the following:)
  getParmSet()->checkParms();
  assert(checkTerms());

  // Set default temperature if not specified in parameter set
  if (!getParmSet()->getDHandler()->isSet(TEMP))
    getParmSet()->getDHandler()->
      setParm(TEMP, options.getDHandler()->getParmVec()[OPT_TEMP]);
  // Create sets of terminals for all cases. We include thermal ports
  // even if they are not needed.
  static const int tvec_idiode[6] = {0,1,0,1,2,3}; // no RS
  static const int tvec_idiodeRS[6] = {4,1,4,1,2,3}; // RS, thermal
  static const int tvec_idiodeRS1[4] = {2,1,2,1}; // RS, no thermal
  static const int tvec_psource[6] = {3,2,0,1,2,3}; // no RS
  static const int tvec_psourceRS[6] = {3,2,4,1,2,3}; //  RS, thermal
  static const int rsterms1[4] = {0,2,0,2}; // no thermal
  static const int rsterms2[4] = {0,4,0,4}; // thermal

  const int* tvec_id;
  const int* tvec_ps;
  // Note: thermal effect of RS not taken into account yet.
  if (dparm[RS] == 0.) {
    setNInternalNodes(0);
    tvec_id = tvec_idiode;
    tvec_ps = tvec_psource;
  } else {
    setNInternalNodes(1);
    LinearVCCS* rs = new RSType(dparm, NULL);
    if (iparm[TPORT]) {
      insert(rs, rsterms2);
      tvec_id = tvec_idiodeRS;
    } else {
      insert(rs, rsterms1);
      tvec_id = tvec_idiodeRS1;
    }
    tvec_ps = tvec_psourceRS;
  }

  GenericVCCS* idiode;
  GenericVCQS* dcharge;
  if (iparm[TPORT]) {
    idiode = new TIntrinsicDiode(dparm, NULL);
    dcharge = new TDiodeCharge(dparm, NULL);
    GenericVCCS* psource = new PowerSource(dparm, NULL); 
    insert(psource, tvec_ps);
  } else {
    idiode = new IntrinsicDiode(dparm, NULL);
    dcharge = new DiodeCharge(dparm, NULL);
  }
  insert(idiode, tvec_id);
  if (getParmSet()->getDHandler()->isSet(CJ0) ||
      getParmSet()->getDHandler()->isSet(TT)) {
    insert(dcharge, tvec_id);
  } else
    delete dcharge;
}

