#include "global.h"
#include "TLinP3.h"
#include <carrot/Options.h>

static ItemInfo TLinP3_info = {
  "tlinp3",
  "Three-terminal physical transmission line",
  "Carlos Chrstoffersen"
  DEFAULT_ADDRESS
};

// Double Parameters
// Make sure enum and ParInfo definitions are in the same order!
enum {K, ALPHA, Z0MAG, FSCALE, TAND, LENGTH, FOPT, MAX_DPARM_INDEX};

//Insert ParmInfo values here
static ParamHandler<double>::ParmInfo TLinP3_dparams[] = { 
  {"k", "Effective dielectric constant", false, 1.},
  {"alpha", "Attenuation (dB/m)", false, .1},
  {"z0mag", "Magnitude of characteristic impedance (ohms)", false, 50.},
  {"fscale", "Scaling frequency for attenuation (Hz)", false, 0.},
  {"tand", "Loss tangent", false, 0.},
  {"length", "Line length (m)", true, 1.},
  {"fopt", "Optimum frequency for discrete approximation", false, 0.}
};

// Uncomment for integer parameters
enum {NSECT, MAX_IPARM_INDEX};
static ParamHandler<int>::ParmInfo TLinP3_iparams[] = {  
  {"nsect", "Discrete approximation with n sections", false, 20},
};

//Library loading mechanism
extern "C" {
  const char * ModuleName(){
    return TLinP3_info.type.c_str();
  }
  ParmSet * newParmSet(const char * model){
    return new ParmSet(model,
		       &TLinP3_info,
		       TLinP3_dparams, MAX_DPARM_INDEX, 
		       TLinP3_iparams, MAX_IPARM_INDEX);
  }
  void delParmSet(ParmSet * p){
    delete p;
    p=0;
  }
  MacroElement * newObject(ParmSet * p, const char * name){
    return new TLinP3(p, name);
  }
  void delObject(MacroElement * e){
    delete e;
    e = 0;
  }
};

#include "LType.h"
#include "RType.h"
#include "CapType.h"
#include "GType.h"

TLinP3::TLinP3(ParmSet *pset, const std::string& name)
  : MacroElement(pset, name, 3), tvec_vcqs(0), tvec_g1(0), tvec_g2(0), 
    tvec_r(0), tvec_cap(0)
{
  //Constructor code goes here (above, 3 is the number of terminals)
}

TLinP3::~TLinP3()
{
  // Delete vectors if at least one of them is allocated
  if (tvec_vcqs) {
    delete [] tvec_vcqs;
    delete [] tvec_g1;
    delete [] tvec_g2;
    delete [] tvec_r;
    delete [] tvec_cap;
  }
}

void TLinP3::init()
{
  const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
  // access to double parameters
  double *dparm = getParmSet()->getDHandler()->getParmVec();
  int *iparm = getParmSet()->getIHandler()->getParmVec();
  // Check required parameters
  getParmSet()->checkParms();
  // Check external terminal connections
  assert(checkTerms());

  // We need at least 2 sections 
  assert(dparm[NSECT] > 1);
  // For gyrators
  assert(scale != 0.);
  // Number of internal nodes
  int int_nodes = 2 * iparm[NSECT] - 1;
  setNInternalNodes(int_nodes);  

  // Here we have to deal with the problem of the internal node
  // vectors (remember to free memory in desttructor):
  tvec_vcqs = new int[4 * iparm[NSECT]];
  tvec_g1 = new int[4 * iparm[NSECT]];
  tvec_g2 = new int[4 * iparm[NSECT]];
  tvec_r = new int[4 * iparm[NSECT]];
  tvec_cap = new int[4 * iparm[NSECT]];

  // Loop for each section 
  const int& iend = iparm[NSECT];
  const int last = iparm[NSECT] - 1;
  int t0 = 0;
  for (int i=0; i<iend; i++) {
    int t3 = 3 + 2*i;
    int t4 = (i==last) ? 1 : t3+1;
    int base = i*4;

    LinearVCQS* ind = new LType(dparm, iparm); // Inductor
    tvec_vcqs[base] = tvec_vcqs[base+2] = t3; 
    tvec_vcqs[base+1] = tvec_vcqs[base+3] = -1; // gnd
    insert(ind, tvec_vcqs+base);
    // Gyrator sources
    LinearVCCS* g1 = new LinearVCCS(scale);
    LinearVCCS* g2 = new LinearVCCS(scale);
    // Add extra node for resistor in series
    tvec_g1[base] = -1; // gnd
    tvec_g1[base+1] = t3; 
    tvec_g1[base+2] = t0; 
    tvec_g1[base+3] = t4;
    tvec_g2[base] = t0;
    tvec_g2[base+1] = t4;
    tvec_g2[base+2] = t3;
    tvec_g2[base+3] = -1; // gnd
    insert(g1, tvec_g1+base);
    insert(g2, tvec_g2+base);
    LinearVCCS* res = new RType(dparm, iparm); // Resistor
    tvec_r[base] = tvec_r[base+2] = t3; 
    tvec_r[base+1] = tvec_r[base+3] = -1; // gnd
    insert(res, tvec_r+base);
    LinearVCQS* cap = new CapType(dparm, iparm); // Capacitor
    // 5 in diagram
    tvec_cap[base] = tvec_cap[base+2] = t4;
    // Always connected to terminal 2
    tvec_cap[base+1] = tvec_cap[base+3] = 2; 
    insert(cap, tvec_cap+base);
    LinearVCCS* gcond = new GType(dparm, iparm); // Conductor
    insert(gcond, tvec_cap+base); // Connected in parallel with cap
    // Update t0;
    t0 = t4;
  }

}

