// This may look like C code, but it is really -*- C++ -*-
//
// Implements a thermal time delayed current source
//
// IMPORTANT: this implementation requires the time step to be fixed
// equal to the time step of the thermal step response. This will be
// changed in the future.
// 

#ifndef ThermalCS_H
#define ThermalCS_H

#include <carrot/TimeDelayedCS.h>

class ThermalCS : public TimeDelayedCS {
public:

  // dparm and iparm values are maintained by the MacroElement
  // (here we assume these are ready to use)
  ThermalCS(const int& ncvolt, const int& maxterms, double* time)
    : TimeDelayedCS(ncvolt), terms(-1), maxterms(maxterms),time(time)
  {
    deltaRth = new double* [getNCVolt()];
    powervec = new double [getNCVolt()*maxterms];
  }

  virtual ~ThermalCS() 
  {
    delete [] deltaRth;
  }

  // Main access function
  virtual void eval(double& f, double* v)
  {
    f = 0.;
    for (int i=0; i<getNCVolt(); i++)
      for (int k=0; k<terms; k++)
	f += powervec[i*maxterms+k] * deltaRth[i][terms-k];
    // Scale factor used in gyrators
    const double &scale = options.getDHandler()->getParmVec()[OPT_SCALE];
    f *= scale * scale;
  }

  // For now do not implement these. In the future a different
  // implementation is possible (with variable time step) and we will
  // need these, but play safe for now.
  virtual void eval_and_deriv(double& f, double* df, double* v)
  {
    assert(false);
  }
  virtual bool paramDeriv(double* df, double *v)
  {
    assert(false);
  }

  // Store port voltages at time t
  virtual void storeVoltages(const double& t, double* v)
  {
    terms++;
    for (int i=0; i<getNCVolt(); i++)
      powervec[i*maxterms + terms] = v[i];
  }

  // Set deltaRth vector for one port
  void setDeltaRthVec(const int& port, double* deltaRthj)
  {
    assert(port < getNCVolt());
    deltaRth[port] = deltaRthj;
  }

private:

  int terms;    // Number of terms in summation
  int maxterms; // Maximum number of terms in the summation
  double* time;
  double** deltaRth;
  double* powervec;
};

#endif
