// This may look like C code, but it is really -*- C++ -*-
//
// IndependentCS is a pure virtual class that handles independent
// current sources:
//
//          1  
//          o        
//          |       
//          |       
//         (^) j(ctime)  
//          |       
//          |     
//          o       
//          2  
//
//
// by Carlos E. Christoffersen

#ifndef IndependentCS_h
#define IndependentCS_h 1

#include <carrot/ElementNode.h>
#include <carrot/Terminal.h>
#include <carrot/InternalCS.h>

class IndependentCS : public ElementNode, public InternalCS
{
public:

  IndependentCS() 
    : ElementNode(2), InternalCS(), row1(-1), row2(-1), deltaT2(0.) {}

  virtual ~IndependentCS() {}

  // This returns the current value. There may be many different
  // implementations. The three time variables are used in multi-time
  // simulations. The first time variable is assumed to be coincident
  // with the `real' time.
  virtual void eval(double& j, 
		    const double& time1,
		    double time2 = -1.,
		    double time3 = -1.) = 0;

  // Initialization: used for taping, error checking, etc.
  // Defaults to nothing.
  // This should be called after any changes in parameters.
  virtual void init() {}

  // We may want to add something that tells us when is there going to
  // be a discontinuity:
  virtual void nextDiscontinuity(double& dtime,
				 const double& time)
  {
    // By default assume no discontinuities are coming
    dtime = 0.;
  }

  // Return the derivative of the current with respect to all double
  // parameters
  virtual bool paramDeriv(double* df, 
			  const double& time1,
			  double time2 = -1.,
			  double time3 = -1.)
  {
    // false means not implemented
    return false;
  }

  // If source has at least 2 time variables: j(t) -> j(t1, t1-deltat2))
  inline void setDeltaT2(const double& dt2)
  {
    deltaT2 = dt2;
  }

  // Functions for quick access to matrix row and column information
  inline const int& getRow1() const
  {
    return row1;
  }

  inline const int& getRow2() const
  {
    return row2;
  }

  inline void setRCinfo()
  {
    row1 = getTerminal(0)->getRC();
    row2 = getTerminal(1)->getRC();
  }

protected:

  // Get offset in second time dimension
  inline const double& getDeltaT2()
  {
    return deltaT2;
  }

private:
  
  int row1, row2;
  // Offset in second time dimension 
  double deltaT2;

};

typedef std::vector<IndependentCS*> IndepCSVector;

#endif


