// This may look like C code, but it is really -*- C++ -*-
//
// LinearVCCS is the class that handles linear voltage-controlled
// current-sources with one control voltage:
//
//          1                        3
//          o                        o       
//          |                               
//          |                        +      
//         (^) g v1                  v1  
//          |                        -      
//          |                             
//          o                        o      
//          2                        4
//
// This is all we need for linear elements. If there are more than one
// controlling voltage we just add 2 or more LinearVCCS in parallel.
//
// by Carlos E. Christoffersen

#ifndef LinearVCCS_h
#define LinearVCCS_h 1

#include <carrot/ElementNode.h>
#include <carrot/Terminal.h>
#include <carrot/InternalCS.h>

class LinearVCCS : public ElementNode, public InternalCS
{
public:

  LinearVCCS(const double& gain) 
    : ElementNode(4), InternalCS(), gain(gain),
      row1(-1), row2(-1), col1(-1), col2(-1) {}

  virtual ~LinearVCCS() {}
    
  // This returns the gain of the VCCS. That is all we really need
  // (besides the terminal connections).
  inline const double getGain() const
  {
    return gain;
  }

  // Set the gain value
  inline void setGain(const double& gain)
  {
    this->gain = gain;
  }

  // Functions for quick access to matrix row and column information
  inline const int& getRow1() const
  {
    return row1;
  }

  inline const int& getRow2() const
  {
    return row2;
  }

  inline const int& getCol1() const
  {
    return col1;
  }

  inline const int& getCol2() const
  {
    return col2;
  }

  inline void setRCinfo()
  {
    row1 = getTerminal(0)->getRC();
    row2 = getTerminal(1)->getRC();
    col1 = getTerminal(2)->getRC();
    col2 = getTerminal(3)->getRC();
  }

  // Return the derivative of g with respect to all double parameters
  virtual bool paramDeriv(double* dg) 
  {
    // False means not implemented
    return false;
  };

private:
  
  double gain;
  int row1, row2, col1, col2;

};

typedef std::vector<LinearVCCS*> LinVCCSVector;

#endif


