// This may look like C code, but it is really -*- C++ -*-
//
// ParmSet is the class to store and manage parameters
//
// by Carlos E. Christoffersen

#ifndef ParmSet_h
#define ParmSet_h 1

#include <carrot/NamedItem.h>
#include <carrot/ElementNode.h>
#include <string>
#include <vector>


class MacroElement;

// Some auxiliary classes and structures first

template <typename T>
class ParamHandler
{
public:  
  
  // Structures to hold parameter information
  struct ParmInfo {
    std::string name;
    std::string comment;
    bool required;       // is this parameter required? (YES = true)
    T defaultValue;      // Default value of parameter
  };

  ParamHandler(ParmInfo* parm_desc, const unsigned& nparms);
  ParamHandler(const ParamHandler<T> & p);

  ~ParamHandler();
  
  inline unsigned getNParms() const 
  {
    return nparms;
  }

  unsigned getIndex(const std::string&) const; 

  void setParm(const unsigned& index, const T &parvalue);

  inline bool isSet(const unsigned& index) const
  {
    assert(index < nparms);
    return parm_set[index];
  }

  void checkParms() const;

  inline const struct ParmInfo& getParmDesc(const unsigned& index) const
  {
    assert(index < nparms);
    return parm_desc[index];
  }


  inline T* getParmVec() 
  {
    return parm;
  }

  inline bool operator==(ParamHandler<T> & rs)
  {
    if(nparms != rs.nparms){
      //Different number of paramters
      return false;
    }

    for(unsigned long int i = 0; i < nparms; ++i){
      if(parm[i] != rs.parm[i]){
        //The actual values are different
        return false;
      }
    }

    return true;
  }

  inline bool operator!=(ParamHandler<T> & rs)
  {
    if( operator==(rs)){
      return false;
    }
    return true;
  }

  ParamHandler<T> & operator=(const ParamHandler<T> & rs);

private:
  
  unsigned nparms;
  ParmInfo* parm_desc;
//  bool* parm_set;
  std::vector<bool> parm_set;
  T* parm;
};

#define DEFAULT_ADDRESS "http://vision.lakeheadu.ca/"

// Structure to hold static information about the item.
struct ItemInfo {
  std::string type;
  std::string description;
  std::string author;
  std::string webaddress;
};

// ********************************************************************
// ParmSet definitions start
// ********************************************************************
class ParmSet : public NamedItem
{

private:
  
  //---------------------------------------
  // Information
  //---------------------------------------
  ItemInfo* i_info;

  //---------------------------------------
  // Parameters
  //---------------------------------------
  ParamHandler<double> dhandler;
  ParamHandler<int> ihandler;
  ParamHandler<std::string> shandler;

  // Vector to all macroelement instances that use this parameter set.
  // I could not use MacroElVector because of the egg and chicken
  // problem.
  ElemVector meVec;

public:
  
  ParmSet(const std::string& name,
	  ItemInfo* i_info, 
	  ParamHandler<double>::ParmInfo *dparam_desc = NULL, 
	  const unsigned numdparms = 0,
	  ParamHandler<int>::ParmInfo *iparam_desc = NULL, 
	  const unsigned numiparms = 0,
	  ParamHandler<std::string>::ParmInfo *sparam_desc = NULL, 
	  const unsigned numsparms = 0);

  // Destructor
  ~ParmSet() {}

  // This is the definition of the different types for parameters
  // Start with a few for now (may be forever).
  enum ParmType {DOUBLE, INT, STRING};

  //---------------------------------------
  // Retrieve Information
  //---------------------------------------
  inline ItemInfo* getInfo() 
  {
    return i_info;
  }  

  // ---------------------------------------
  // Parameter stuff
  // ---------------------------------------

  bool askAbout(ParmType& type, unsigned& index, 
		const std::string& parName) const; 
  
  //The following cannot be inlined because their definition is in the header,
  // but they depend on optionally compiled and emitted code in .cc files.
  //  Solution - don't make the code inline and stick it in ParmSet.cc 
  // Functions to set parameters (one per type)
  void setParm(const unsigned& index, const double& parvalue);
  void setParm(const unsigned& index, const int& parvalue);
  void setParm(const unsigned& index, const std::string& parvalue);

  // Check element parameters. 
  // This may not be really necessary since we can call this
  // individually on each handler.
  void checkParms() const;

  void print();

  // The following are used mainly by MacroElements
  inline ParamHandler<double> *getDHandler()
  {
    return &dhandler;
  }
  inline ParamHandler<int> *getIHandler()
  {
    return &ihandler;
  }
  inline ParamHandler<std::string> *getSHandler()
  {
    return &shandler;
  }

  inline void addME(MacroElement* me)
  {
    meVec.push_back((ElementNode*)me);
  }

  inline void removeME(MacroElement* me)
  {
    ElementNode* en = (ElementNode*)me;
    int size = meVec.size();
    for (int i(0); i<size; i++)
      if (meVec[i] == en) {
	// Exchange with last element
	meVec[i] = meVec[size-1];
	meVec.pop_back();
	break;
      }
  }

  inline int getMEVecSize()
  {
    return meVec.size();
  }

  inline MacroElement* getME(const int& i)
  {
    return (MacroElement*)(meVec[i]);
  }

  bool operator==(ParmSet & rs);
  bool operator!=(ParmSet & rs);

  ParmSet & operator=(const ParmSet & rs);
};

typedef std::vector<ParmSet*> ParmSetVector;

#endif


