// This may look like C code, but it is really -*- C++ -*-
//
// This class is to store and handle a real sparse square matrix using
// the UMFPack library.
//
// Remember: UMFPack uses zero-indexing (C convention)
// 

#ifndef UMFPackZI_H
#define UMFPackZI_H

#include <complex>

typedef std::complex<double> double_complex;

class UMFPackZI 
{
public:

  // iSize is the initial estimation of the matrix size.
  UMFPackZI(const int& nrcs, const int& iSize)
    : nrcs(nrcs), flag(false), factor_flag(false), symbolic_flag(false),
      triplet_g_size(0), triplet_lin_size(0), rowvec(), colvec(), valuevec(),
      Symbolic(NULL), Numeric(NULL)
  {
    rowvec.reserve(iSize);
    colvec.reserve(iSize);
    valuevec.reserve(iSize);
  }

  ~UMFPackZI();

  // This is the main method to add elements to the matrix. A row or
  // column value of -1 (ground row or column) indicates the value
  // will not be added.
  inline void addElement(const int& row, const int& col, const double_complex& value)
  {
    if (row >= 0 && col >= 0) {
      // Add one triplet
      rowvec.push_back(row);
      colvec.push_back(col);
      valuevec.push_back(value);
    }
  }

  // Set the mark for the part of the triplet that is not erased when
  // calling clearCapNumeric(). By the default the mark is set to zero
  // (all elements erased).
  inline void setGMark()
  {
    triplet_g_size = rowvec.size();
  }

  // Set the mark for the part of the triplet that is not erased when
  // calling clearNonlinearNumeric(). By the default the mark is set
  // to zero (all elements erased).
  inline void setLinMark()
  {
    triplet_lin_size = rowvec.size();
  }

  // Build compressed-column matrix
  void buildCompressed();

  // Preorder matrix (umfpack_*_symbolic, required before factor)
  int symbolic();

  // Factor sparse matrix (required before solve)
  int factor();

  // Backward-substitution
  int solve(double_complex* x, double_complex* rhs);

  // Backward-substitution of transposed matrix
  int solveTransposed(double_complex* x, double_complex* rhs);

  // Clear nonlinear part of triplet and numeric structure (at end of
  // iteration)
  void clearNonlinearNumeric();

  // Clear nonlinear part and capacitors of triplet and numeric
  // structure (when time step is changed, for example)
  void clearCapNumeric();

  // Remove memory associated with compressed matrix (and
  // symbolic). Use this if the matrix structure needs to be changed.
  void clearCompressed();

  // Reset the matrix
  void reset();

  // For debugging purposes: Prints the sparse matrix directly from
  // the triplet form.
  void printTriplet();

private:

  int nrcs;
  bool flag, factor_flag, symbolic_flag;
  int triplet_g_size;
  int triplet_lin_size;
  // These are the vectors to hold the matrix in triplet form  
  std::vector<int> rowvec, colvec; 
  std::vector<double_complex> valuevec;
  // Variables needed for UMFPack
  void *Symbolic, *Numeric;
  // Compressed column format vectors (and Map)
  int *colPtr, *rowIdx, *Map;
  double_complex *valA;
  int *Wi;
  double *W;
};

#endif
