// Main function when carrot runs in batch mode

#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <carrot/Circuit.h>
#include <carrot/Analysis.h>
#include "analysis/AnalysisLoader.h"
#include <carrot/ObjectLoader.h>
#include <carrot/ParmSet.h>
#include <carrot/Options.h>
#include "tokenizer.h"

#include <strings.h>
#include <stdlib.h>


void parse(Circuit &cir, 
	   std::map<std::string, ParmSet *> &analysisParameters,
	   AnalysisMap &analysisLibraries, std::istream *file_p)
{
  // Shared libraries already loaded
  ModuleMap &circuitElements = cir.getModuleMap();

  std::map<std::string, Terminal* > terminals;
  std::map<std::string, MacroElement *> macroElements;
  std::map<std::string, ParmSet *> models;
  std::vector<ParmSet *> parameters;


  //When we loop through the incomming input, we will need some
  //storage locations
  std::string inputLine;
  bool lineHandled = false; //Used to allow filtering by different handlers
  bool doneInput = false;

  unsigned long int lineNum = 0;

  //Create ground terminal

  terminals["gnd"] = cir.getGroundTerm();
  terminals["0"] = cir.getGroundTerm();
	
  while( ! file_p->eof() && !doneInput ){
    std::getline((*file_p), inputLine);
    ++lineNum;
    lineHandled = false;
		
    if(!lineHandled && tokenizeSkipLine::isSkipLine(inputLine) ){
      //We should skip this line and not do anything with it
      lineHandled = true;
    }


    if(!lineHandled && tokenizeEndOfFile::isEndOfFile(inputLine) ){
      lineHandled = true;
      doneInput = true;
    }

    if( !lineHandled && 
	tokenizeModuleDefinition::isModuleDefinition(inputLine)) {
      lineHandled = true;
      
      tokenizeModuleDefinition mod(inputLine);

      //Check to ensure that we support this element type
      if(circuitElements.find(mod.getElementName()) == circuitElements.end()){
	//We don't support this element type
	std::cerr << "Element type\""
		  << mod.getElementName() 
		  << "\" not supported at input line: "
		  << lineNum << std::endl;
	exit(1);
      }

      //Check to make sure that we have not already used this model name;
      if(models.find(mod.getInstanceName()) != models.end()){
	std::cerr << "Model name \"";
	std::cerr << mod.getInstanceName() << 
	  "\" has already been defined.  Line :";
	std::cerr << lineNum << std::endl;
	exit(1);
      }


      //Check to make sure that a reference model already exists if specified
      if(mod.getRefModelName().length() > 0 
	 && models.find(mod.getRefModelName()) == models.end()) {
	std::cerr << "Base model name \"";
	std::cerr << mod.getRefModelName() << 
	  "\" has not been defined.  Line :";
	std::cerr << lineNum << std::endl;
	exit(1);
      }
			
      ParmSet * pset = circuitElements[mod.getElementName()].newParmSet(mod.getInstanceName().c_str());

      if(mod.getRefModelName().length() > 0) {
	*pset = *models[mod.getRefModelName()];
      }

      std::map<std::string, std::string> parms = mod.getParameters();
      std::map<std::string, std::string>::iterator it = parms.begin();

      unsigned int index;
      ParmSet::ParmType pType;

      //Process the parameters and add them to the mix

      while(it != parms.end()){
	//Insure good parameter name string
	if(it->first.length() == 0){
	  std::cerr << "Invalid token at the end of line ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//Insure good parameter value string
	if(it->second.length() == 0){
	  std::cerr << "Parameter \"" << it->first << "\"";
	  std::cerr << "Requires a value at line: ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//How do I handle invalid parameter names?  
	if (!pset->askAbout(pType, index, it->first.c_str())) {
	  std::cout << "\nWarning *** Unknown parameter: \"";
	  std::cout << it->first << "\" at line " << lineNum << std::endl;
	}

	if(pType == ParmSet::DOUBLE){
	  pset->setParm(index, atof(it->second.c_str()));
	}else if(pType == ParmSet::INT){
	  pset->setParm(index, atoi(it->second.c_str()));
	}else if(pType == ParmSet::STRING){
	  pset->setParm(index, it->second.c_str());
	}else{
	  std::cout << "\nWarning *** I don't know how to handle parameter \"";
	  std::cout << it->first << "\"" << std::endl;
	}
	++it;
      }

      //Even if this model duplicates the parent, we will hang on to
      //it.  This allows us to change the values later for whatever
      //reason we might want
      parameters.push_back(pset);
      models[mod.getInstanceName()] = pset;
    }	

    if( !lineHandled && tokenizeAnalysisType::isAnalysisType(inputLine) ){
      tokenizeAnalysisType item(inputLine);
      lineHandled = true;

      if(analysisLibraries.find(item.getAnalysisName()) == analysisLibraries.end()){
        //We don't support this analysis type
        std::cerr << "Analysis type \"";
        std::cerr << item.getAnalysisName();
        std::cerr << "\" not supported at input line: ";
        std::cerr << lineNum << std::endl;
        exit(1);
      }

      //Check to see if this analysis type has already been defined
      if(analysisParameters.find(item.getAnalysisName()) != analysisParameters.end()){
        //We have already defined this analysis type
        std::cerr << "Analysis type \"";
        std::cerr << item.getAnalysisName();
        std::cerr << "\" already in use at input line: ";
        std::cerr << lineNum << std::endl;
        exit(1);
      }

      ParmSet * pset = analysisLibraries[item.getAnalysisName()].newParmSet( std::string(item.getAnalysisName() + "_analysis").c_str());

      std::map<std::string, std::string> parms = item.getParameters();
      std::map<std::string, std::string>::iterator it = parms.begin();

      ParmSet::ParmType pType;
      unsigned int index;

      //Process the parameters and add them to the mix

      while(it != parms.end()){
	//Insure good parameter name string
	if(it->first.length() == 0){
	  std::cerr << "Invalid token at the end of line ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//Insure good parameter value string
	if(it->second.length() == 0){
	  std::cerr << "Parameter \"" << it->first << "\"";
	  std::cerr << "Requires a value at line: ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//How do I handle invalid parameter names?  
	if (!pset->askAbout(pType, index, it->first.c_str())) {
	  std::cout << "\nWarning *** Unknown parameter: \"";
	  std::cout << it->first << "\" at line " << lineNum << std::endl;
	}

	if(pType == ParmSet::DOUBLE){
	  pset->setParm(index, atof(it->second.c_str()));
	}else if(pType == ParmSet::INT){
	  pset->setParm(index, atoi(it->second.c_str()));
	}else if(pType == ParmSet::STRING){
	  pset->setParm(index, it->second.c_str());
	}else{
	  std::cout << "\nWarning *** I don't know how to handle parameter \"";
	  std::cout << it->first << "\"" << std::endl;
	}
	++it;
      }			

      //Now we add this to the map until we are ready to being analysis
      analysisParameters[item.getAnalysisName()] = pset;

    }

    // Handle global circuit options. Added by CEC 2/6/2005
    if( !lineHandled && tokenizeOptionsType::isOptionsType(inputLine) ){
      tokenizeOptionsType item(inputLine);
      lineHandled = true;

      std::map<std::string, std::string> parms = item.getParameters();
      std::map<std::string, std::string>::iterator it = parms.begin();

      ParmSet::ParmType pType;
      unsigned int index;

      //Process the parameters and add them to the mix
      while(it != parms.end()){
	//Insure good parameter name string
	if(it->first.length() == 0){
	  std::cerr << "Invalid token at the end of line ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//Insure good parameter value string
	if(it->second.length() == 0){
	  std::cerr << "Parameter \"" << it->first << "\"";
	  std::cerr << "Requires a value at line: ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//How do I handle invalid parameter names?  
	if (!options.askAbout(pType, index, it->first.c_str())) {
	  std::cout << "\nWarning *** Unknown parameter: \"";
	  std::cout << it->first << "\" at line " << lineNum << std::endl;
	}

	if(pType == ParmSet::DOUBLE){
	  options.setParm(index, atof(it->second.c_str()));
	}else if(pType == ParmSet::INT){
	  options.setParm(index, atoi(it->second.c_str()));
	}else if(pType == ParmSet::STRING){
	  options.setParm(index, it->second.c_str());
	}else{
	  std::cout << "\nWarning *** I don't know how to handle parameter \"";
	  std::cout << it->first << "\"" << std::endl;
	}
	++it;
      }			

      // Nothing else needs to be done since the parameters are
      // already inserted.
    }



    //Deal with individual circuit elements
    if(!lineHandled && tokenizeCircuitItem::isCircuitItem(inputLine) ){
      tokenizeCircuitItem item(inputLine);

      lineHandled = true;

      //Check to make sure that we support this element type
      if(circuitElements.find(item.getElementName()) == circuitElements.end()){
	//We don't support this element type
	std::cerr << "Element type \"";
	std::cerr << item.getElementName() 
		  << "\" not supported at input line: ";
	std::cerr << lineNum << std::endl;
	exit(1);
      }
			
      if(item.getElementName().length() < 1){
	std::cerr << "Instance name too short" << std::endl;
	exit (1);
      }
			
      if(macroElements.find(item.getInstanceName()) != macroElements.end() ){
	std::cerr << "Instance name of \"";
	std::cerr << item.getInstanceName() << "\" already in use at line: ";
	std::cerr << lineNum << std::endl;
	exit(1);
      }


      //Check to make sure that a reference model already exists if specified
      if(item.getRefModelName().length() > 0 
	 && models.find(item.getRefModelName()) == models.end())
      {
	std::cerr << "Base model name \"";
	std::cerr << item.getRefModelName() 
		  << "\" has not been defined.  Line :";
	std::cerr << lineNum << std::endl;
	exit(1);
      }

      //Create parameter set
      ParmSet * pset = circuitElements[item.getElementName()].newParmSet(item.getInstanceName().c_str());

      if(item.getRefModelName().length() > 0){
	*pset = *models[item.getRefModelName()];
      }

      std::map<std::string, std::string> parms = item.getParameters();
      std::map<std::string, std::string>::iterator it = parms.begin();

      ParmSet::ParmType pType;
      unsigned int index;

      //Process the parameters and add them to the mix

      while(it != parms.end()){
	//Insure good parameter name string
	if(it->first.length() == 0){
	  std::cerr << "Invalid token at the end of line ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//Insure good parameter value string
	if(it->second.length() == 0){
	  std::cerr << "Parameter \"" << it->first << "\"";
	  std::cerr << "Requires a value at line: ";
	  std::cerr << lineNum << std::endl;
	  exit(1);
	}

	//How do I handle invalid parameter names?  
	if (!pset->askAbout(pType, index, it->first.c_str())) {
	  std::cout << "\nWarning *** Unknown parameter: \"";
	  std::cout << it->first << "\" at line " << lineNum << std::endl;
	}

	if(pType == ParmSet::DOUBLE){
	  pset->setParm(index, atof(it->second.c_str()));
	}else if(pType == ParmSet::INT){
	  pset->setParm(index, atoi(it->second.c_str()));
	}else if(pType == ParmSet::STRING){
	  pset->setParm(index, it->second.c_str());
	}else{
	  std::cout << "\nWarning *** I don't know how to handle parameter \"";
	  std::cout << it->first << "\"" << std::endl;
	}
	++it;
      }			

      //Add parameters to the maps
	
      //Check to see if we can avoid the creation of another parameter set

      if(item.getRefModelName().length() > 0 && *pset 
	 == *models[item.getRefModelName()]) {
	delete pset;
	pset = models[item.getRefModelName()];
      }
      else {
	parameters.push_back(pset);
      }

      //Create the object proper
      MacroElement * mElement;
      mElement = circuitElements[item.getElementName()].newObject(pset, item.getInstanceName().c_str());
      macroElements[item.getInstanceName()] = mElement;

      //Now we need to attach terminals
      //For each terminal for the device..

      // std::cout << "Number of terminals for device: " << mElement->getNTerms() << std::endl;

      if(item.getTerminals().size() != mElement->getNTerms()){
	std::cerr << "Element type \"";
	std::cerr << item.getElementName() 
		  << "\" has " << mElement->getNTerms();
	std::cerr << " terminals which must be connected.  Instead found ";
	std::cerr << item.getTerminals().size() << " at line: ";
	std::cerr << lineNum << std::endl;
	exit(1);
      }


      for(unsigned int i = 0; i < item.getTerminals().size() ; ++i){
	//Check to see if the terminal name is already in use
	if(terminals.find(item.getTerminals()[i]) == terminals.end() ){
	  //No - create the terminal
	  terminals[item.getTerminals()[i]] = new Terminal(item.getTerminals()[i].c_str());
	  mElement->link(i, terminals[item.getTerminals()[i]]);
	}else{
	  //Yes - we can use the existing terminal
	  mElement->link(i, terminals[item.getTerminals()[i]]);
	}
      }



    }
  }

  //Now we can attach everything to the circuit

  for(std::vector<ParmSet *>::iterator i = parameters.begin(); 
      i != parameters.end(); ++i) {
    cir.addParmSet(*i);
  }

  for(std::map<std::string, 
	MacroElement *>::iterator i = macroElements.begin(); 
      i !=macroElements.end(); ++i) {
    cir.addElement(i->second);
  }

  for(std::map<std::string, Terminal* >::iterator i = terminals.begin(); 
      i!=terminals.end(); ++i){
    if(i->first !="gnd" && i->first !="0") { 
      //Adding the ground terminal again causes problems
      cir.addTerminal(i->second);
    }
  }


  //Free up all possible memory
  parameters.clear();
  macroElements.clear();
  terminals.clear();


}
